#include "gsmhacking.h"

struct bone **alloc_bones(unsigned short n) {
  return (struct bone **) alloc_lots(n, "bone", sizeof(struct bone *), (void *(*)()) alloc_bone);
}

struct bone *alloc_bone() {
  return (struct bone *) alloc_one("bone", sizeof(struct bone));
}

void cleanup_bones(unsigned short num_bones, struct bone **bones, unsigned short num_frames) {
  unsigned short i;

  if (bones) {
    for (i = 0; i < num_bones; i++) {
      if (bones[i]->animated) {
        cleanup_quaternions(num_frames, bones[i]->orientations);
        cleanup_matrices(num_frames, bones[i]->locals);
        cleanup_matrices(num_frames, bones[i]->worlds);
      }
      TRACE2("freeing bone name %s\n", bones[i]->name);
      if (bones[i]->name) free(bones[i]->name);
    }
   free_structs("bone", num_bones, (void *) bones);
  }
}

int find_bones(struct gsm *gsm, struct skeleton *skeleton) {
  unsigned short i;
  unsigned int pos;
  struct bone *bone, *parent;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think bones start at 0x%04x\n", pos);

  /* allocate a big enough struct */
  skeleton->bones = alloc_bones(skeleton->num_bones);
  if (! skeleton->bones) return 1;

  /* fill in the bones */
  for (i = 0; i < skeleton->num_bones; i++) {
    bone = skeleton->bones[i];
    bone->name = 0;
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 1;
    bone->offset = pos;
    if (parse_bone(gsm, i, bone, skeleton->num_bones)) {
      ERROR("expected bone id %u but got null\n", i);
      return 2;
    }
    if (bone->links_to == BONE_FIRST_LINK) bone->parent = 0;
    else bone->parent = skeleton->bones[bone->links_to];
    if (calculate_bone_matrices(bone)) return 3;
  }

  return 0;
}

int parse_bone(struct gsm *gsm, unsigned short num, struct bone *bone, unsigned short total) {
  /* find id */
  if (read_short(gsm->fd, &bone->number)) return 1;
  TRACE("this is bone #%u of %u\n", num, total);

  /* find name length */
  if (read_short(gsm->fd, &bone->name_length)) {
    ERROR("can't read length of bone name");
    return 2;
  }
  TRACE("name is of length %u\n", bone->name_length);

  /* read name */
  bone->name = (char *) malloc((sizeof(char) * bone->name_length + 1));
  if (! bone->name) {
    ERROR("no memory for bone name");
    return 3;
  }
  if (read_arbitrary(gsm->fd, bone->name, bone->name_length)) {
    ERROR("can't read name of bone");
    return 4;
  }
  (bone->name)[bone->name_length] = '\0';

  /* which bone does this connect to? */
  if (read_short(gsm->fd, &bone->links_to)) {
    ERROR("can't read bone link");
    return 5;
  }
  if (read_vector(gsm->fd, &bone->position)) {
    ERROR("can't read bone locale co-ordinates");
    return 6;
  }
  if (read_quaternion(gsm->fd, &bone->q)) {
    ERROR("can't read bone orientation");
    return 7;
  }

  TRACE("bone %u at 0x%04x: %s (% 0.4f, % 0.4f, % 0.4f), Q(% 0.4f, (% 0.4f, % 0.4f, % 0.4f))\n", bone->number, bone->offset, bone->name, bone->position.x, bone->position.y, bone->position.z, bone->q.w, bone->q.x, bone->q.y, bone->q.z);

  return 0;
}

struct bone **make_swt_bones() {
  unsigned short i;
  struct bone **bones;

  bones = alloc_bones(SWT_BIPED_NUM_BONES);
  if (! bones) {
    ERROR("out of memory for bones in new SWT skeleton\n");
    return 0;
  }

  for (i = 0; i < SWT_BIPED_NUM_BONES; i++) {
    if (make_swt_bone(i, bones[i])) {
      ERROR("can't make bone %u\n", i);
      cleanup_bones(SWT_BIPED_NUM_BONES, bones, 0);
      return 0;
    }
    if (bones[i]->links_to == BONE_FIRST_LINK) bones[i]->parent = 0;
    else bones[i]->parent = bones[bones[i]->links_to];
    if (calculate_bone_matrices(bones[i])) {
      ERROR("can't calculate bone %u matrices\n", i);
      return 0;
    }
  }

  return bones;
}

int make_swt_bone(unsigned short num, struct bone *bone) {
  bone->offset = 0x0fc0;
  bone->number = num;
  TRACE1("making new bone %u\n", bone->number);
  bone->name = (char *) alloc_name(get_biped_bone_name(GSM_TYPE_SWT, bone->number), &bone->name_length);
  if (! bone->name) {
    ERROR("can't set bone #%02u name\n", bone->number);
    return 1;
  }
  TRACE2("setting bone %u co-ordinates\n", bone->number);
  /* stuff is stored in a contiguous block of memory so what the hell, let's go for it (very ugly though) */
  /* XXX: naughty, shouldn't mix sizeof() and explicit number */
  memmove(&bone->links_to, swt_bone_coords[bone->number], sizeof(short));
  memmove(&bone->position.x, swt_bone_coords[bone->number] + 2, sizeof(float));
  memmove(&bone->position.y, swt_bone_coords[bone->number] + 6, sizeof(float));
  memmove(&bone->position.z, swt_bone_coords[bone->number] + 10, sizeof(float));
  memmove(&bone->q.x, swt_bone_coords[bone->number] + 14, sizeof(float));
  memmove(&bone->q.y, swt_bone_coords[bone->number] + 18, sizeof(float));
  memmove(&bone->q.z, swt_bone_coords[bone->number] + 22, sizeof(float));
  memmove(&bone->q.w, swt_bone_coords[bone->number] + 26, sizeof(float));
  TRACE2("bone #%02u new name: %s\n", bone->number, bone->name);

  return 0;
}

struct bone **make_kid_bones() {
  unsigned short i;
  struct bone **bones;

  bones = alloc_bones(KID_BIPED_NUM_BONES);
  if (! bones) {
    ERROR("out of memory for bones in new KID skeleton\n");
    return 0;
  }

  for (i = 0; i < KID_BIPED_NUM_BONES; i++) {
    if (make_kid_bone(i, bones[i])) {
      ERROR("can't make bone %u\n", i);
      cleanup_bones(KID_BIPED_NUM_BONES, bones, 0);
      return 0;
    }
    if (bones[i]->links_to == BONE_FIRST_LINK) bones[i]->parent = 0;
    else bones[i]->parent = bones[bones[i]->links_to];
    if (calculate_bone_matrices(bones[i])) {
      ERROR("can't calculate bone %u matrices\n", i);
      return 0;
    }
  }

  return bones;
}

int make_kid_bone(unsigned short num, struct bone *bone) {
  bone->offset = 0x0fc0;
  bone->number = num;
  TRACE1("making new bone %u\n", bone->number);
  bone->name = (char *) alloc_name(get_biped_bone_name(GSM_TYPE_KID, bone->number), &bone->name_length);
  if (! bone->name) {
    ERROR("can't set bone #%02u name\n", bone->number);
    return 1;
  }
  TRACE2("setting bone %u co-ordinates\n", bone->number);
  /* stuff is stored in a contiguous block of memory so what the hell, let's go for it (very ugly though) */
  /* XXX: naughty, shouldn't mix sizeof() and explicit number */
  memmove(&bone->links_to, kid_bone_coords[bone->number], sizeof(short));
  memmove(&bone->position.x, kid_bone_coords[bone->number] + 2, sizeof(float));
  memmove(&bone->position.y, kid_bone_coords[bone->number] + 6, sizeof(float));
  memmove(&bone->position.z, kid_bone_coords[bone->number] + 10, sizeof(float));
  memmove(&bone->q.x, kid_bone_coords[bone->number] + 14, sizeof(float));
  memmove(&bone->q.y, kid_bone_coords[bone->number] + 18, sizeof(float));
  memmove(&bone->q.z, kid_bone_coords[bone->number] + 22, sizeof(float));
  memmove(&bone->q.w, kid_bone_coords[bone->number] + 26, sizeof(float));
  TRACE2("bone #%02u new name: %s\n", bone->number, bone->name);

  return 0;
}

struct bone **make_tac_bones() {
  unsigned short i;
  struct bone **bones;

  bones = alloc_bones(TAC_BIPED_NUM_BONES);
  if (! bones) {
    ERROR("out of memory for bones in new TAC skeleton\n");
    return 0;
  }

  for (i = 0; i < TAC_BIPED_NUM_BONES; i++) {
    if (make_tac_bone(i, bones[i])) {
      ERROR("can't make bone %u\n", i);
      cleanup_bones(TAC_BIPED_NUM_BONES, bones, 0);
      return 0;
    }
    if (bones[i]->links_to == BONE_FIRST_LINK) bones[i]->parent = 0;
    else bones[i]->parent = bones[bones[i]->links_to];
    if (calculate_bone_matrices(bones[i])) {
      ERROR("can't calculate bone %u matrices\n", i);
      return 0;
    }
  }

  return bones;
}

int make_tac_bone(unsigned short num, struct bone *bone) {
  bone->offset = 0x0fc0;
  bone->number = num;
  TRACE1("making new bone %u\n", bone->number);
  bone->name = (char *) alloc_name(get_biped_bone_name(GSM_TYPE_TAC, bone->number), &bone->name_length);
  if (! bone->name) {
    ERROR("can't set bone #%02u name\n", bone->number);
    return 1;
  }
  TRACE2("setting bone %u co-ordinates\n", bone->number);
  /* stuff is stored in a contiguous block of memory so what the hell, let's go for it (very ugly though) */
  memmove(&bone->links_to, tac_bone_coords[bone->number], sizeof(short));
  memmove(&bone->position.x, tac_bone_coords[bone->number] + 2, sizeof(float));
  memmove(&bone->position.y, tac_bone_coords[bone->number] + 6, sizeof(float));
  memmove(&bone->position.z, tac_bone_coords[bone->number] + 10, sizeof(float));
  memmove(&bone->q.x, tac_bone_coords[bone->number] + 14, sizeof(float));
  memmove(&bone->q.y, tac_bone_coords[bone->number] + 18, sizeof(float));
  memmove(&bone->q.z, tac_bone_coords[bone->number] + 22, sizeof(float));
  memmove(&bone->q.w, tac_bone_coords[bone->number] + 26, sizeof(float));
  TRACE2("bone #%02u new name: %s\n", bone->number, bone->name);

  return 0;
}

struct bone **make_bf1_bones() {
  unsigned short i;
  struct bone **bones;

  bones = alloc_bones(BF1_BIPED_NUM_BONES);
  if (! bones) {
    ERROR("out of memory for bones in new SWT skeleton\n");
    return 0;
  }

  for (i = 0; i < BF1_BIPED_NUM_BONES; i++) {
    if (make_bf1_bone(i, bones[i])) {
      ERROR("can't make bone %u\n", i);
      cleanup_bones(BF1_BIPED_NUM_BONES, bones, 0);
      return 0;
    }
    if (bones[i]->links_to == BONE_FIRST_LINK) bones[i]->parent = 0;
    else bones[i]->parent = bones[bones[i]->links_to];
    if (calculate_bone_matrices(bones[i])) {
      ERROR("can't calculate bone %u matrices\n", i);
      return 0;
    }
  }

  return bones;
}

int make_bf1_bone(unsigned short num, struct bone *bone) {
  bone->offset = 0x0fc0;
  bone->number = num;
  TRACE1("making new bone %u\n", bone->number);
  bone->name = (char *) alloc_name(get_biped_bone_name(GSM_TYPE_SWT, bone->number), &bone->name_length);
  if (! bone->name) {
    ERROR("can't set bone #%02u name\n", bone->number);
    return 1;
  }
  TRACE2("setting bone %u co-ordinates\n", bone->number);
  /* stuff is stored in a contiguous block of memory so what the hell, let's go for it (very ugly though) */
  /* XXX: naughty, shouldn't mix sizeof() and explicit number */
  memmove(&bone->links_to, bf1_bone_coords[bone->number], sizeof(short));
  memmove(&bone->position.x, bf1_bone_coords[bone->number] + 2, sizeof(float));
  memmove(&bone->position.y, bf1_bone_coords[bone->number] + 6, sizeof(float));
  memmove(&bone->position.z, bf1_bone_coords[bone->number] + 10, sizeof(float));
  memmove(&bone->q.x, bf1_bone_coords[bone->number] + 14, sizeof(float));
  memmove(&bone->q.y, bf1_bone_coords[bone->number] + 18, sizeof(float));
  memmove(&bone->q.z, bf1_bone_coords[bone->number] + 22, sizeof(float));
  memmove(&bone->q.w, bf1_bone_coords[bone->number] + 26, sizeof(float));
  TRACE2("bone #%02u new name: %s\n", bone->number, bone->name);

  return 0;
}

int write_bones(struct gsm *gsm, struct skeleton *skeleton) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing bones: %u\n", skeleton->num_bones);
  if (write_short(gsm->out_fd, &skeleton->num_bones)) {
    ERROR("can't write number of bones");
    return 2;
  }

  for (i = 0; i < skeleton->num_bones; i++) {
    TRACE("writing bone id %u\n", i);
    if (write_short(gsm->out_fd, &skeleton->bones[i]->number)) {
      ERROR("can't write bone %u id", i);
      return 3;
    }
    TRACE("writing bone name \"%s\" of length %u\n", skeleton->bones[i]->name, skeleton->bones[i]->name_length);
    if (write_short(gsm->out_fd, &skeleton->bones[i]->name_length)) {
      ERROR("can't write bone %u name length", i);
      return 4;
    }
    if (write_arbitrary(gsm->out_fd, skeleton->bones[i]->name, skeleton->bones[i]->name_length)) {
      ERROR("can't write bone %u name", i);
      return 5;
    }
    if (write_short(gsm->out_fd, &skeleton->bones[i]->links_to)) {
      ERROR("can't write bone %u link", i);
      return 6;
    }
    if (write_vector(gsm->out_fd, &skeleton->bones[i]->position)) {
      ERROR("can't write bone %u co-ordinates", i);
      return 7;
    }
    if (write_quaternion(gsm->out_fd, &skeleton->bones[i]->q)) {
      ERROR("can't write bone %u co-ordinates", i);
      return 7;
    }
  }

  return 0;
}

/* assume only one skeleton */
char *get_bone_by_id(struct gsm *gsm, unsigned short bone_id) {
  if (! gsm->num_skeletons || ! gsm->skeletons || ! gsm->skeletons[0]) {
    ERROR("no skeletons in GSM\n");
    return "bogus";
  }
  if (bone_id >= gsm->skeletons[0]->num_bones) return 0;
  return gsm->skeletons[0]->bones[bone_id]->name;
}

char *get_bone_link(struct gsm *gsm, unsigned short bone_id) {
  if (! gsm->num_skeletons || ! gsm->skeletons || ! gsm->skeletons[0] || ! gsm->skeletons[0]->bones) {
    ERROR("no skeletons in GSM\n");
    return "bogus";
  }
  if (bone_id > gsm->skeletons[0]->num_bones) return 0;
  if (gsm->skeletons[0]->bones[bone_id]->links_to == BONE_FIRST_LINK) return "root";
  return get_bone_by_id(gsm, gsm->skeletons[0]->bones[bone_id]->links_to);
}

int get_bone_by_name(struct gsm *gsm, char *name) {
  unsigned short i;

  if (! gsm) {
    ERROR("no GSM!\n");
    return -2;
  }
  if (! name) {
    ERROR("no name to search for\n");
    return -3;
  }
  if (! gsm->num_skeletons || ! gsm->skeletons || ! gsm->skeletons[0] || ! gsm->skeletons[0]->bones) {
    ERROR("no skeletons in GSM\n");
    return -4;
  }

  for (i = 0; i < gsm->skeletons[0]->num_bones; i++) {
    /* case-insensitive match */
    if (! gsm->skeletons[0]->bones[i]) {
      ERROR("bone %u is null\n", i);
      return -5;
    }
    if (str_equiv(gsm->skeletons[0]->bones[i]->name, name)) return i;
  }

  ERROR("bone \"%s\" not found in skeleton\n", name);
  return -1;
}

int parse_orientation(struct gsm *gsa, unsigned short num, quaternion_t *orientation, unsigned short frame) {
  if (! gsa || ! gsa->fd) {
    ERROR("no file descriptor for gsa\n");
    return 1;
  }

  if (read_quaternion(gsa->fd, orientation)) {
    ERROR("can't read orientation\n");
    return 2;
  }

  TRACE("frame %u orientation %u (bone %d): Q(% 0.4f, (% 0.4f, % 0.4f, % 0.4f))\n", frame, num, get_bone_id_by_animated_bone_entry(gsa, num), orientation->w, orientation->x, orientation->y, orientation->z);

  return 0;
}

int write_orientation(struct gsm *gsa, unsigned short num, quaternion_t *orientation, unsigned short frame) {
  if (! gsa || ! gsa->out_fd) {
    ERROR("no file descriptor for gsa\n");
    return 1;
  }

  if (write_quaternion(gsa->out_fd, orientation)) {
    ERROR("can't write orientation rot_x");
    return 2;
  }

  return 0;
}

/* populate matrices using data we just read */
int calculate_bone_matrices(struct bone *bone) {
  matrix_t *m;
  vector_t *v;

  if (! bone) {
    ERROR("no bone!\n");
    return 1;
  }

  /* construct local matrix from rotation */
  if (quaternion_to_matrix(&bone->local, &bone->q)) {
    ERROR("can't convert bone \"%s\" orientation to local rotation matrix\n", bone->name);
    return 3;
  }
  /* and fill in the translation part of the matrix using the (x,y,z) vector */
  if (set_matrix_translation(&bone->local, &bone->position)) {
    ERROR("can't set translation part of local matrix for bone \"%s\"\n", bone->name);
    return 4;
  }

  /* root bone world matrix IS the local matrix */
  if (bone->links_to == BONE_FIRST_LINK) {
    copy_matrix(bone->world, bone->local);
    return 0;
  }

  m = multiply_matrices(&bone->parent->world, &bone->local);
  if (! m) {
    ERROR("can't convert bone \"%s\" co-ordinate system\n", bone->name);
    return 6;
  }
  copy_matrix(&bone->world, m);
  free(m);

  /* this really is very inefficient -> find world position */
  v = multiply_vector_by_matrix(&bone->position, &bone->world);
  if (! v) {
    ERROR("can't find bone \"%s\"'s position in new co-ordinate system\n", bone->name);
    return 7;
  }
  v->x += bone->parent->world[TX];
  v->y += bone->parent->world[TY];
  v->z += bone->parent->world[TZ];
  set_matrix_translation(&bone->world, v);
  free(v);

  return 0;
}

int describe_bones(struct gsm *gsm) {
  unsigned short i, x, y;
  struct bone *bone;
  euler_t e;
  unsigned int packed;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  for (i = 0; i < gsm->skeletons[0]->num_bones; i++) {
    bone = gsm->skeletons[0]->bones[i];
    if (! gsm_verbose) {
      SAY("%s %02u at 0x%04x: %s, linked to %s\n", (bone->animated) ? "Animated bone" : "Bone", i, bone->offset, get_bone_by_id(gsm, i), get_bone_link(gsm, i));
      continue;
    }
    SAY("%s %02u at 0x%04x: %s\n", (bone->animated) ? "Animated bone" : "Bone", i, bone->offset, get_bone_by_id(gsm, i));
    SAY("  Position: (% 0.4f, % 0.4f, % 0.4f)\n", bone->world[TX], bone->world[TY], bone->world[TZ]);
    SAY("  Rotation: Q(% 0.4f, (% 0.4f, % 0.4f, % 0.4f))\n", bone->q.w, bone->q.x, bone->q.y, bone->q.z);
    if (gsm_verbose > 1) {
      SAY("  Local position: (% 0.4f, % 0.4f, % 0.4f)\n", bone->local[TX], bone->local[TY], bone->local[TZ]);
      //SAY("  World rotation: Q(% 0.4f, (% 0.4f, % 0.4f, % 0.4f))\n", bone->world_q.w, bone->world_q.x, bone->world_q.y, bone->world_q.z);
    }
    if (gsm_verbose > 1) {
      if (! quaternion_to_euler(&e, &bone->q)) {
        SAY("  Euler: (% 0.4f, % 0.4f, % 0.4f)d (% 0.4f, % 0.4f, % 0.4f)r\n", radians_to_degrees(e[0]), radians_to_degrees(e[1]), radians_to_degrees(e[2]), e[0], e[1], e[2]);
      }
      if (gsm_verbose > 2) {
        SAY("  Local   | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->local[R0], bone->local[R3], bone->local[R6], bone->local[TX]);
        SAY("  matrix: | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->local[R1], bone->local[R4], bone->local[R7], bone->local[TY]);
        SAY("          | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->local[R2], bone->local[R5], bone->local[R8], bone->local[TZ]);

        SAY("\n");

        SAY("  World   | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->world[R0], bone->world[R3], bone->world[R6], bone->world[TX]);
        SAY("  matrix: | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->world[R1], bone->world[R4], bone->world[R7], bone->world[TY]);
        SAY("          | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", bone->world[R2], bone->world[R5], bone->world[R8], bone->world[TZ]);
      }
    }
    if (bone->links_to == BONE_FIRST_LINK) {
      SAY("  Root bone\n");
    }
    else {
      SAY("  Links to: %02u %s\n", bone->links_to, get_bone_link(gsm, i));
    }

    if (gsm->is_animation) continue;

    /* assignments */
    packed = get_assignments_by_bone(gsm, i);
    if (packed) {
      x = packed >> 8;
      y = packed % 256;
      if (x == 1) {
        SAY("  One vertex: %02u\n", y);
      }
      else {
        SAY("  Assigned vertices: %u, from %02u to %02u\n", x, y, y + x - 1);
      }
    }
    else {
      SAY("  No vertex assignments\n");
    }
    SAY("\n");
  }

  return 0;
}
