#include "gsmhacking.h"

int read_gsa_file(struct gsm *gsa) {
  if (! gsa || ! gsa->filename) {
    ERROR("no filename to open!\n");
    return -1;
  }
  gsa->type |= GSM_TYPE_ANIMATION;
  gsa->is_animation = 1;
  gsa->fd = open_read(gsa->filename);
  if (! gsa->fd) {
    ERROR("can't open file %s!\n", gsa->filename);
    return -2;
  }
  DEBUG("opened %s on descriptor %p\n", gsa->filename, gsa->fd);
  if (check_magic(gsa, GSA_MAGIC)) {
    DEBUG("%s isn't a gsa file\n", gsa->filename);
    fclose(gsa->fd);
    return 1;
  }
  set_basename(gsa->filename, 0, &gsa->basename, &gsa->basename_length);
  INFO("%s is an animation file\n", gsa->basename);

  return 0;
}

int parse_gsa(struct gsm *gsa) {
  if (! gsa || ! gsa->fd) {
    ERROR("parse_gsa() called with no gsa\n");
    return 1;
  }

  if (find_animation_stuff(gsa)) return 1;
  if (find_skeletons(gsa)) return 2;
  if (find_frames(gsa)) return 3;

  return 0;
}

void classify_gsa(struct gsm *gsa) {
  if (! gsa || ! gsa->filename || ! gsa->version || ! gsa->num_skeletons) {
    ERROR("classify_gsa() called with no gsa\n");
    return;
  }

  DEBUG("checking GSA animation\n");
  gsa->type %= GSM_TYPE_ANIMATION_LIMITED_BONES;

  DEBUG("checking number of skeletons\n");
  if (gsa->num_skeletons != 1) {
    WARN("GSA should have one skeleton only\n");
  }

  DEBUG("checking number of bones\n");
  switch (gsa->skeletons[0]->num_bones) {
    case 2:
      DEBUG("2 bones => weapon, rat or fish\n");
      DEBUG("checking skeleton name(!)\n");
      if (str_equiv(gsa->skeletons[0]->name, "Rat_Root")) {
        TRACE("%s => rat\n", gsa->skeletons[0]->name);
        gsa->type = GSM_TYPE_RAT;
      }
      else if (str_equiv(gsa->skeletons[0]->name, "FISH_Root")) {
        DEBUG("%s => fish\n", gsa->skeletons[0]->name);
        gsa->type = GSM_TYPE_FISH;
      }
      else if (str_equiv(gsa->skeletons[0]->name, "Rifle_Root")) {
        DEBUG("%s => weapon\n", gsa->skeletons[0]->name);
        gsa->type = GSM_TYPE_GUN;
      }
      else if (str_equiv(gsa->skeletons[0]->name, "Bone01")) {
        DEBUG("%s => unknown, may be casing\n", gsa->skeletons[0]->name);
        gsa->type = GSM_TYPE_GUN;
      }
      else {
        DEBUG("%s => unknown, may be weapon\n", gsa->skeletons[0]->name);
        gsa->type = GSM_TYPE_GUN;
      }
      break;
    case 34:
      DEBUG("34 bones => BF1 or FHx character\n");
      gsa->type = GSM_TYPE_BF1;
      break;
    case 37:
      DEBUG("37 bones => TAC or MHx character\n");
      gsa->type = GSM_TYPE_TAC;
      break;
    case 40:
      DEBUG("40 bones => KID\n");
      gsa->type = GSM_TYPE_KID;
      break;
    case 52:
      DEBUG("52 bones => SWT character\n");
      gsa->type = GSM_TYPE_SWT;
      break;
    default:
      WARN("unrecognised number of bones in GSA: %u\n", gsa->skeletons[0]->num_bones);
      break;
  }

  DEBUG("checking number of exported bones\n");
  if (gsa->skeletons[0]->num_animated_bones != gsa->skeletons[0]->num_bones) {
    DEBUG("%u out of %u animated\n", gsa->skeletons[0]->num_animated_bones, gsa->skeletons[0]->num_bones);
    gsa->type |= GSM_TYPE_ANIMATION_LIMITED_BONES;
  }

  gsa->type |= GSM_TYPE_ANIMATION;
}

int describe_gsa(struct gsm *gsa) {
  unsigned short i, j;
  char *tmp;

  if (! gsa || ! gsa->filename) {
    ERROR("no GSA to describe\n");
    return 1;
  }

  if (describe_skeletons(gsa)) return 2;
  if (describe_frames(gsa)) return 3;

  return 0;
}

void summarise_gsa(struct gsm *gsa) {
  if (! gsa || ! gsa->filename) {
    ERROR("no GSA to summarise\n");
    return;
  }

  SAY("Summary for %s\n", gsa->basename);
  if (gsa->skeletons[0]->num_bones == gsa->skeletons[0]->num_animated_bones) {
    SAY("Bones: %u\n", gsa->skeletons[0]->num_bones);
  }
  else {
    SAY("Bones: %u (%u animated)\n", gsa->skeletons[0]->num_bones, gsa->skeletons[0]->num_animated_bones);
  }
  SAY("Frames: %u\n", gsa->num_frames);
  SAY("Events: %u\n", gsa->skeletons[0]->num_events);
}

void show_gsa(struct gsm *gsa) {
  char *description;
  unsigned int type;

  if (! gsa || ! gsa->filename) {
    ERROR("no GSM to describe\n");
    return;
  }

  if (gsa->type & GSM_TYPE_ANIMATION_LIMITED_BONES) description = "limited animation";
  else description = "animation";

  type = gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES;

  if (gsm_verbose) {
    switch (type) {
      case GSM_TYPE_UNKNOWN:
        SAY("%s: unknown %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_GUN:
        SAY("%s: rifle %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_SWT:
        SAY("%s: SWAT officer %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_TAC:
        SAY("%s: male suspect %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_MHx:
        SAY("%s: male hostage %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_BF1:
        SAY("%s: female suspect %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_FHx:
        SAY("%s: female hostage %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_KID:
        SAY("%s: child hostage %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_RAT:
        SAY("%s: rat %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_FISH:
        SAY("%s: fish %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_OBJ:
        SAY("%s: (throwable) object %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_MAG:
        SAY("%s: casing or magazine %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_PIS:
        SAY("%s: pistol %s\n", gsa->filename, description);
        break;
      case GSM_TYPE_TOOL:
        SAY("%s: tool %s\n", gsa->filename, description);
        break;
      default:
        ERROR("GSM type %u is unrecognised; this is a bug\n", gsa->type);
    }
  }
  else {
    switch (type) {
      case GSM_TYPE_UNKNOWN: SAY("unknown %s\n", description); break;
      case GSM_TYPE_GUN: SAY("rifle %s\n", description); break;
      case GSM_TYPE_SWT: SAY("SWT %s\n", description); break;
      case GSM_TYPE_TAC: SAY("TAC %s\n", description); break;
      case GSM_TYPE_MHx: SAY("MHx %s\n", description); break;
      case GSM_TYPE_BF1: SAY("BF1 %s\n", description); break;
      case GSM_TYPE_FHx: SAY("FHx %s\n", description); break;
      case GSM_TYPE_KID: SAY("KID %s\n", description); break;
      case GSM_TYPE_RAT: SAY("RAT %s\n", description); break;
      case GSM_TYPE_FISH: SAY("FISH %s\n", description); break;
      case GSM_TYPE_OBJ: SAY("throwable %s\n", description); break;
      case GSM_TYPE_MAG: SAY("casing/magazine %s\n", description); break;
      case GSM_TYPE_PIS: SAY("pistol %s\n", description); break;
      case GSM_TYPE_TOOL: SAY("tool %s\n", description); break;
      default: ERROR("unrecognised type %u; this is a bug\n", gsa->type);
    }
  }
}

int write_gsa_data(struct gsm *gsa) {
  if (! gsa || ! gsa->out_fd) {
    ERROR("no gsa to write!\n");
    return 1;
  }

  /* do the business */
  if (write_magic(gsa)) return 5;
  if (write_animation_stuff(gsa)) return 7;
  if (write_skeletons(gsa)) return 8;
  if (write_frames(gsa)) return 9;

  return 0;
}

/* is this a "selected bones only" animation */
int find_animated_bones(struct gsm *gsa, struct skeleton *skeleton) {
  unsigned short i;
  unsigned short magic;

  if (! gsa || ! gsa->fd) {
    ERROR("no fd for gsa!\n");
    return 6; /* 1-5 used by find_skeletons() */
  }
  
  if (read_short(gsa->fd, &magic)) {
    ERROR("can't read marker after bone data\n");
    return 7;
  }
  if (magic != GSA_END_OF_BONE_STREAM) {
    WARN("expected %u after bone data but got %u\n", GSA_END_OF_BONE_STREAM, magic);
  }

  if (read_short(gsa->fd, &skeleton->num_animated_bones)) {
    ERROR("can't read number of animated bones in this animation\n");
    return 8;
  }

  if (! skeleton->num_animated_bones) {
    INFO("animation applies to all %u bones\n", skeleton->num_bones);
    skeleton->num_animated_bones = skeleton->num_bones;
    for (i = 0; i < skeleton->num_bones; i++) skeleton->bones[i]->animated = 1;
    return 0;
  }
  else if (skeleton->num_animated_bones == 1) {
    INFO("animation applies to only one bone in the skeleton\n");
  }
  else {
    INFO("animation applies to %u bones in the skeleton\n", skeleton->num_animated_bones);
  }

  skeleton->animated_bones = (unsigned short **) alloc_lots(skeleton->num_animated_bones, "animated_bone", sizeof(unsigned short *), (void *(*)()) alloc_short);
  if (! skeleton->animated_bones) {
    ERROR("out of memory for animated bone list");
    return 9;
  }

  for (i = 0; i < skeleton->num_animated_bones; i++) {
    if (read_short(gsa->fd, &magic)) {
      ERROR("can't read animated bone ID %u\n", i);
      return 10;
    }
    if (magic >= skeleton->num_bones) {
      ERROR("aniamted bone index (%u) out of range (%u bones)\n", magic, skeleton->num_bones);
      return 11;
    }
    skeleton->bones[magic]->animated = 1;
    *skeleton->animated_bones[i] = magic;
  }

  return 0;
}

/* is this a "selected bones only" animation */
int write_animated_bones(struct gsm *gsa, struct skeleton *skeleton) {
  unsigned short i;
  unsigned short magic;

  if (! gsa || ! gsa->out_fd) {
    ERROR("no fd for gsa!\n");
    return 7; /* 1-6 used by write_skeletons() */
  }
  
  magic = GSA_END_OF_BONE_STREAM;
  if (write_short(gsa->out_fd, &magic)) {
    ERROR("can't write marker after bone data\n");
    return 8;
  }

  /* no need to write the list if all bones are animated */
  if (skeleton->num_animated_bones == skeleton->num_bones) {
    DEBUG("marking all %u bones for animation\n", gsa->skeletons[0]->num_animated_bones);
    magic = 0;
    if (write_short(gsa->out_fd, &magic)) {
      ERROR("can't write zero animated bones\n");
      return 9;
    }
    return 0;
  }

  DEBUG("writing list of animated bones: %u\n", gsa->skeletons[0]->num_animated_bones);
  if (write_short(gsa->out_fd, &skeleton->num_animated_bones)) {
    ERROR("can't write number of animated bones in this animation\n");
    return 9;
  }

  for (i = 0; i < skeleton->num_bones; i++) {
    if (! skeleton->bones[i]->animated) continue;
    TRACE("bone %2u %s is animated\n", i, get_biped_bone_name(GSM_TYPE_SWT, i));
    if (write_short(gsa->out_fd, &i)) {
      ERROR("can't write animated bone ID %u\n", i);
      return 10;
    }
  }

  return 0;
}
