#include "gsmhacking.h"

struct gsm **alloc_gsms(unsigned short n) {
  return (struct gsm **) alloc_lots(n, "gsm", sizeof(struct gsm *), (void *(*)()) alloc_gsm);
}

struct gsm *alloc_gsm() {
  return (struct gsm *) alloc_one("gsm", sizeof(struct gsm));
}

/*
   open the file and parse the gsm magic
   gsm->fd is set to the new file descriptor
   returns 0 on success
*/
int read_gsm_file(struct gsm *gsm) {
  if (! gsm || ! gsm->filename) {
    ERROR("no filename to open!\n");
    return -1;
  }
  gsm->fd = open_read(gsm->filename);
  if (! gsm->fd) {
    ERROR("can't open file %s!\n", gsm->filename);
    return -2;
  }
  DEBUG("opened %s on descriptor %p\n", gsm->filename, gsm->fd);
  if (check_magic(gsm, GSM_MAGIC)) {
    DEBUG("%s isn't a gsm file\n", gsm->filename);
    fclose(gsm->fd);
    return 1;
  }
  set_basename(gsm->filename, 0, &gsm->basename, &gsm->basename_length);

  return 0;
}

int write_gsm_file(struct gsm *gsm) {
  if (! gsm || (! gsm->filename && ! gsm->option->fd)) {
    ERROR("no gsm to write!\n");
    return 1;
  }

  if (open_gsm_file(gsm)) return 2;
  if (gsm->type & GSM_TYPE_ANIMATION) {
    if (write_gsa_data(gsm)) return 3;
  }
  else if (write_gsm_data(gsm)) return 3;

  return 0;
}

int open_gsm_file(struct gsm *gsm) {
  char *dirname;
  unsigned short dirname_length;
  int need_free;

  if (! gsm || (! gsm->filename && ! gsm->option->fd)) {
    ERROR("no gsm to write!\n");
    return 1;
  }

  /* write to stdout? */
  if (gsm->option->fd) gsm->out_fd = gsm->option->fd;
  else {
    /* use existing filename if no output filename was given */
    gsm->outfile = gsm->option->outfile;
    if (! gsm->outfile) {
      gsm->outfile = gsm->filename;
      /* was there a file prefix? */
      if (gsm->option->prefix) gsm->outfile = add_prefix(gsm->outfile, gsm->option->prefix);
      if (gsm->option->suffix) gsm->outfile = add_suffix(gsm->outfile, gsm->option->suffix);
      if (! gsm->outfile) {
        ERROR("gsm->outfile got mashed\n");
        return 2;
      }
    }
    set_basename(gsm->outfile, 0, &gsm->out_basename, &gsm->out_basename_length);
    /* create tmp file */
    gsm->tmpfile = (char *) malloc(sizeof(char) * 24 + strlen(gsm->outfile));
    if (! gsm->tmpfile) {
      ERROR("out of memory for gsm temporary output file");
      return 3;
    }

    dirname = get_dirname(gsm->outfile, &dirname_length, &need_free);
    if (! dirname) {
      ERROR("can't get outfile dirname\n");
      return 4;
    }

    DEBUG("dirname_length: %u\n", dirname_length);
    if (dirname_length) {
      *(dirname + dirname_length - 1) = PATHSEP;
      *(dirname + dirname_length) = '\0';
    }
    else {
      /*free(dirname);*/ /* not needed? */
      need_free = 0;
      dirname = "";
    }
    if (snprintf(gsm->tmpfile, 24 + strlen(gsm->outfile), "%s%u.%u.%s", dirname, time(0), getpid(), gsm->out_basename) == -1) {
      ERROR("snprintf() failed writing scratch file\n");
      return 5;
    }
    if (need_free) free(dirname);

    gsm->out_fd = open_write(gsm->tmpfile);
    if (! gsm->out_fd) {
      ERROR("can't open tmpfile %s!\n", gsm->tmpfile);
      return 6;
    }
    DEBUG("opened tmpfile %s on descriptor %p\n", gsm->tmpfile, gsm->out_fd);
  }

  return 0;
}

int write_gsm_data(struct gsm *gsm) {
  if (! gsm || ! gsm->out_fd) {
    ERROR("no gsm to write!\n");
    return 1;
  }

  /* do the business */
  if (write_magic(gsm)) return 5;
  if (write_skeletons(gsm)) return 6;
  if (write_assignments(gsm)) return 7;
  if (write_positions(gsm)) return 8;
  if (write_blended_positions(gsm)) return 9;
  if (write_texture_coords(gsm)) return 10;
  if (write_vertices(gsm)) return 11;
  if (write_triangles(gsm)) return 12;
  if (write_triangle_groups(gsm)) return 13;
  if (write_shared_triangle_lists(gsm)) return 14;
  if (write_textures(gsm)) return 15;
  if (write_materials(gsm)) return 16;
  if (write_grips(gsm)) return 17;
  if (write_grippers(gsm)) return 18;

  return 0;
}

int link_gsm_file(struct gsm *gsm) {
  if (! gsm || ! gsm->out_fd) {
    ERROR("gsm has no open file descriptor\n");
    return 1;
  }

  if (gsm->option->fd);
  else {
    DEBUG("closing descriptor %p for %s\n", gsm->out_fd, gsm->tmpfile);
    if (fclose(gsm->out_fd)) {
      ERROR("failed to close file descriptor %p\n", gsm->out_fd);
      return 2;
    }
    gsm->out_fd = 0;
  
    DEBUG("linking real output filename %s\n", gsm->outfile);
#ifdef _MSC_VER
    if (str_equals(gsm->filename, gsm->outfile)) {
      if (unlink(gsm->outfile)) {
        ERROR("failed to unlink original file %s prior to replacement\n", gsm->outfile);
      }
    }
#endif
    if (rename(gsm->tmpfile, gsm->outfile)) {
      ERROR("failed to create output file %s\n", gsm->outfile);
      ERROR("GSM data can be found in temp file %s\n", gsm->tmpfile);
      return 3;
    }
  
    INFO("%s written successfully\n", gsm->outfile);
  }
  if (gsm->outfile == gsm->filename) gsm->outfile = 0;
  return 0;
}

int close_gsm_file(struct gsm *gsm) {
  int ret;

  if (! gsm || ! gsm->fd) {
    ERROR("gsm has no open file descriptor\n");
    return -1;
  }
  DEBUG("closing descriptor %p for %s\n", gsm->fd, gsm->filename);
  ret = fclose(gsm->fd);
  gsm->fd = 0;

  return ret;
}

int write_magic(struct gsm *gsm) {
  if (! gsm || ! gsm->out_fd) {
    ERROR("gsm has no open file descriptor\n");
    return 1;
  }

  if (gsm->type & GSM_TYPE_ANIMATION) {
    if (write_arbitrary(gsm->out_fd, GSA_MAGIC, GSA_MAGIC_LEN)) {
      ERROR("can't write magic");
      return 2;
    }
  }
  else {
    if (write_arbitrary(gsm->out_fd, GSM_MAGIC, GSM_MAGIC_LEN)) {
      ERROR("can't write magic");
      return 2;
    }
  }
  if (write_short(gsm->out_fd, &gsm->version)) {
    ERROR("can't write version");
    return 4;
  }
  return 0;
}

/* check that the first few octets of the file match the .gsm signature */
int check_magic(struct gsm *gsm, char *magic) {
  int i;

  if (read_arbitrary(gsm->fd, gsm_buffer, GSM_MAGIC_LEN)) {
    ERROR("error reading magic");
    return -1;
  }
  for (i = 0; i < GSM_MAGIC_LEN; i++) {
    if (gsm_buffer[i] != magic[i]) {
      DEBUG("magic[%d]: 0x%02x (expected 0x%02x)\n", i, (unsigned char) gsm_buffer[i], (unsigned char) magic[i]);
      return -1;
    }
    TRACE("magic[%d]: 0x%02x (OK)\n", i, (unsigned char) magic[i]);
  }
  if (read_short(gsm->fd, &gsm->version)) {
    ERROR("can't get version info");
    return -1;
  }
  if (gsm->type & GSM_TYPE_ANIMATION) {
    DEBUG("gsa exporter version %u.%u\n", gsm->version >> 8, gsm->version % 16);
  }
  else {
    DEBUG("gsm exporter version %u.%u\n", gsm->version >> 8, gsm->version % 16);
  }

  return 0;
}

/* free all memory used by the gsm */
void cleanup_gsm(struct gsm *gsm) {
  if (! gsm) return;
  if (gsm->num_skeletons) cleanup_skeletons(gsm->num_skeletons, gsm->skeletons, gsm->num_frames);
  if (gsm->num_assignments) cleanup_assignments(gsm->num_assignments, gsm->assignments);
  if (gsm->num_positions) cleanup_positions(gsm->num_positions, gsm->positions);
  if (gsm->num_blended_positions) cleanup_blended_positions(gsm->num_blended_positions, gsm->blended_positions);
  if (gsm->num_texture_coords) cleanup_texture_coords(gsm->num_texture_coords, gsm->texture_coords);
  if (gsm->num_vertices) cleanup_vertices(gsm->num_vertices, gsm->vertices);
  if (gsm->num_triangles) cleanup_triangles(gsm->num_triangles, gsm->triangles);
  if (gsm->num_triangle_groups) cleanup_triangle_groups(gsm->num_triangle_groups, gsm->triangle_groups);
  if (gsm->num_shared_triangle_lists) cleanup_shared_triangle_lists(gsm->num_shared_triangle_lists, gsm->shared_triangle_lists);
  if (gsm->num_textures) cleanup_textures(gsm->num_textures, gsm->textures);
  if (gsm->num_materials) cleanup_materials(gsm->num_materials, gsm->materials);
  if (gsm->num_grips) cleanup_grips(gsm->num_grips, gsm->grips);
  if (gsm->num_grippers) cleanup_grippers(gsm->num_grippers, gsm->grippers);
  if (gsm->num_frames) cleanup_frames(gsm->num_frames, gsm->frames);
  if (gsm->filename) free(gsm->filename);
  if (gsm->outfile) free(gsm->outfile);
  if (gsm->tmpfile) free(gsm->tmpfile);
  cleanup_option(gsm->option);
  TRACE2("freeing GSM at %p\n", gsm);
  free(gsm);
}

/* this is it, the big function! */
int parse_gsm(struct gsm *gsm) {
  if (! gsm || ! gsm->fd) {
    ERROR("parse_gsm() called with no gsm\n");
    return 1;
  }

  if (find_skeletons(gsm)) return 1;
  if (find_assignments(gsm)) return 2;
  if (find_positions(gsm)) return 3;
  if (find_blended_positions(gsm)) return 4;
  if (find_texture_coords(gsm)) return 5;
  if (find_vertices(gsm)) return 6;
  if (find_triangles(gsm)) return 7;
  if (find_triangle_groups(gsm)) return 8;
  if (find_shared_triangle_lists(gsm)) return 9;
  if (find_textures(gsm)) return 10;
  if (find_materials(gsm)) return 11;
  if (find_grips(gsm)) return 12;
  if (find_grippers(gsm)) return 13;

  if (gsm->unmatched_materials) gsm->need_optimise = 1;
  if (gsm->type == GSM_TYPE_PIS) gsm->need_optimise = 1;

  return 0;
}

void classify_gsm(struct gsm *gsm) {
  if (! gsm || ! gsm->filename || ! gsm->version || ! gsm->num_skeletons) {
    ERROR("classify_gsm() called with no gsm\n");
    return;
  }

  DEBUG("checking GSM version\n");
  if (gsm->version >> 8 != 4 || gsm->version % 16 != 5) {
    WARN("don't know about GSM exporters other than 4.5\n");
  }

  DEBUG("checking number of skeletons\n");
  if (gsm->num_skeletons != 1) {
    WARN("GSM should have one skeleton only\n");
  }

  DEBUG("checking number of bones\n");
  switch (gsm->skeletons[0]->num_bones) {
    case 2:
      DEBUG("2 bones => weapon, rat or fish\n");
      DEBUG("checking skeleton name(!)\n");
      if (str_equiv(gsm->skeletons[0]->name, "Rat_Root")) {
        TRACE("%s => rat\n", gsm->skeletons[0]->name);
        gsm->type = GSM_TYPE_RAT;
      }
      else if (str_equiv(gsm->skeletons[0]->name, "FISH_Root")) {
        DEBUG("%s => fish\n", gsm->skeletons[0]->name);
        gsm->type = GSM_TYPE_FISH;
      }
      else if (str_equiv(gsm->skeletons[0]->name, "Rifle_Root")) {
        DEBUG("%s => weapon\n", gsm->skeletons[0]->name);
        gsm->type = GSM_TYPE_GUN;
      }
      else if (str_equiv(gsm->skeletons[0]->name, "Bone01")) {
        DEBUG("%s => unknown, may be casing\n", gsm->skeletons[0]->name);
        gsm->type = GSM_TYPE_GUN;
      }
      else {
        DEBUG("%s => unknown, may be weapon\n", gsm->skeletons[0]->name);
        gsm->type = GSM_TYPE_GUN;
      }
      break;
    case 34:
      DEBUG("34 bones => BF1 or FHx character\n");
      gsm->type = GSM_TYPE_BF1 | GSM_TYPE_FHx;
      break;
    case 37:
      DEBUG("37 bones => TAC or MHx character\n");
      gsm->type = GSM_TYPE_TAC | GSM_TYPE_MHx;
      break;
    case 40:
      DEBUG("40 bones => KID\n");
      gsm->type = GSM_TYPE_KID;
      break;
    case 52:
      DEBUG("52 bones => SWT character\n");
      gsm->type = GSM_TYPE_SWT;
      break;
    default:
      WARN("unrecognised number of bones in GSM: %u\n", gsm->skeletons[0]->num_bones);
      break;
  }

  DEBUG("checking number of grips\n");
  if (gsm->num_grips == 2) {
    switch (gsm->type) {
      case GSM_TYPE_GUN:
        DEBUG("%u => rifle\n", gsm->num_grips);
        break;
      default:
        DEBUG("%u => FIXME, what is this?\n", gsm->num_grips);
        break;
    }
  }
  else if (gsm->num_grips == 1) {
    switch (gsm->type) {
      case GSM_TYPE_GUN:
        DEBUG("%u => pistol\n", gsm->num_grips);
        gsm->type = GSM_TYPE_PIS;
        break;
      default:
        DEBUG("%u => doesn't tell us anything new\n", gsm->num_grips);
        break;
    }
  }

  DEBUG("checking number of grippers\n");
  if (gsm->num_grippers == 3) {
    switch (gsm->type) {
      case GSM_TYPE_BF1 | GSM_TYPE_FHx:
        DEBUG("%u => female suspect\n", gsm->num_grippers);
        gsm->type = GSM_TYPE_BF1;
        break;
      case GSM_TYPE_TAC | GSM_TYPE_MHx:
        DEBUG("%u => male suspect\n", gsm->num_grippers);
        gsm->type = GSM_TYPE_TAC;
        break;
      case GSM_TYPE_GUN:
        DEBUG("%u => rifle\n", gsm->num_grippers);
        break;
      default:
        DEBUG("%u => doesn't tell us anything new\n", gsm->num_grippers);
        break;
    }
  }
  else if (gsm->num_grippers == 0) {
    switch (gsm->type) {
      case GSM_TYPE_BF1 | GSM_TYPE_FHx:
        DEBUG("none => female hostage\n");
        gsm->type = GSM_TYPE_FHx;
        break;
      case GSM_TYPE_TAC | GSM_TYPE_MHx:
        DEBUG("none => male hostage\n");
        gsm->type = GSM_TYPE_MHx;
        break;
      default:
        DEBUG("none => doesn't tell us anything new\n", gsm->num_grippers);
        break;
    }
  }
  else {
    DEBUG("%u => FIXME, what is this?\n", gsm->num_grippers);
  }

  if (gsm->type == GSM_TYPE_GUN || gsm->type == GSM_TYPE_PIS) {
    if (get_material_by_name(gsm, MUZZLE) >= 0) {
      DEBUG("GSM has muzzle material and two bones => weapon\n");
    }
    else {
      DEBUG("no muzzle => magazine, casing, throwable or tool\n");
      if (gsm->num_grips == 3) {
        DEBUG("3 grips => throwable\n");
        gsm->type = GSM_TYPE_OBJ;
      }
      if (gsm->num_grips == 2) {
        DEBUG("2 grips => tool\n");
        gsm->type = GSM_TYPE_TOOL;
      }
      else {
        DEBUG("no grips => magazine or casing\n");
        gsm->type = GSM_TYPE_MAG;
      }
    }
  }
}

void show_gsm(struct gsm *gsm) {
  if (! gsm || ! gsm->filename) {
    ERROR("no GSM to describe\n");
    return;
  }

  if (gsm_verbose) {
    switch (gsm->type) {
      case GSM_TYPE_UNKNOWN:
        SAY("%s: unknown\n", gsm->filename);
        break;
      case GSM_TYPE_GUN:
        SAY("%s: rifle\n", gsm->filename);
        break;
      case GSM_TYPE_SWT:
        SAY("%s: SWAT officer\n", gsm->filename);
        break;
      case GSM_TYPE_TAC:
        SAY("%s: male suspect\n", gsm->filename);
        break;
      case GSM_TYPE_MHx:
        SAY("%s: male hostage\n", gsm->filename);
        break;
      case GSM_TYPE_BF1:
        SAY("%s: female suspect\n", gsm->filename);
        break;
      case GSM_TYPE_FHx:
        SAY("%s: female hostage\n", gsm->filename);
        break;
      case GSM_TYPE_KID:
        SAY("%s: child hostage\n", gsm->filename);
        break;
      case GSM_TYPE_RAT:
        SAY("%s: rat\n", gsm->filename);
        break;
      case GSM_TYPE_FISH:
        SAY("%s: fish\n", gsm->filename);
        break;
      case GSM_TYPE_OBJ:
        SAY("%s: (throwable) object\n", gsm->filename);
        break;
      case GSM_TYPE_MAG:
        SAY("%s: casing or magazine\n", gsm->filename);
        break;
      case GSM_TYPE_PIS:
        SAY("%s: pistol\n", gsm->filename);
        break;
      case GSM_TYPE_TOOL:
        SAY("%s: tool\n", gsm->filename);
        break;
      default:
        ERROR("GSM type %u is unrecognised; this is a bug\n", gsm->type);
    }
  }
  else {
    switch (gsm->type) {
      case GSM_TYPE_UNKNOWN: SAY("unknown\n"); break;
      case GSM_TYPE_GUN: SAY("rifle\n"); break;
      case GSM_TYPE_SWT: SAY("SWT\n"); break;
      case GSM_TYPE_TAC: SAY("TAC\n"); break;
      case GSM_TYPE_MHx: SAY("MHx\n"); break;
      case GSM_TYPE_BF1: SAY("BF1\n"); break;
      case GSM_TYPE_FHx: SAY("FHx\n"); break;
      case GSM_TYPE_KID: SAY("KID\n"); break;
      case GSM_TYPE_RAT: SAY("RAT\n"); break;
      case GSM_TYPE_FISH: SAY("FISH\n"); break;
      case GSM_TYPE_OBJ: SAY("throwable\n"); break;
      case GSM_TYPE_MAG: SAY("casing/magazine\n"); break;
      case GSM_TYPE_PIS: SAY("pistol\n"); break;
      case GSM_TYPE_TOOL: SAY("tool\n"); break;
      default: ERROR("unrecognised type %u; this is a bug\n", gsm->type);
    }
  }
}

void show_polycount(struct gsm *gsm) {
  if (! gsm || ! gsm->filename) {
    ERROR("no GSM for polygon count\n");
    return;
  }

  if (gsm_verbose) SAY("%s: %u polygons\n", gsm->basename, gsm->num_triangles);
  else SAY("%u\n", gsm->num_triangles);
}

void summarise_gsm(struct gsm *gsm) {
  if (! gsm || ! gsm->filename) {
    ERROR("no GSM to summarise\n");
    return;
  }

  SAY("Summary for %s\n", gsm->basename);
  SAY("Bones: %u\n", gsm->skeletons[0]->num_bones);
  SAY("Vertex to bone assignments: %u\n", gsm->num_assignments);
  SAY("Positions (vertices): %u\n", gsm->num_positions);
  SAY("Blended positions: %u (length %d)\n", gsm->num_blended_positions, gsm->blended_position_stream_length);
  SAY("Effective vertices: %u\n", gsm->num_vertices);
  SAY("Texture co-ordinates: %u\n", gsm->num_texture_coords);
  SAY("Triangles: %u\n", gsm->num_triangles);
  SAY("Triangle groups: %u\n", gsm->num_triangle_groups);
  SAY("Shared triangle lists: %u (length %d)\n", gsm->num_shared_triangle_lists, gsm->shared_triangle_list_length * 2 - 2);
  SAY("Textures: %u\n", gsm->num_textures);
  SAY("Materials: %u\n", gsm->num_materials);
  SAY("Grips: %u\n", gsm->num_grips);
  SAY("Grippers: %u\n", gsm->num_grippers);
}

int describe_gsm(struct gsm *gsm) {
  unsigned short i;
  struct material *material;
  struct map *map;

  if (! gsm || ! gsm->filename) {
    ERROR("no GSM to describe\n");
    return 1;
  }

  /* skin-o-matic? */
  if (gsm->option->describe == DESCRIBE_SKIN_O_MATIC) {
    /* write material_id <tab> map_type <tab> transparency <tab> texture_name */
    for (i = 0; i < gsm->num_materials; i++) {
      material = gsm->materials[i];
      if (! material->num_maps) continue;
      map = material->maps[0];
      SAY("%u\t%s\t%d\t%s\n", i, get_map_type_name(map->map_type), (int) (map->texture_id));
    }
    return 0;
  }

  if (gsm_verbose) {
    SAY("GSkinMesh: %s\n", gsm->filename);
    SAY("Skeletons in mesh: %u\n\n", gsm->num_skeletons);
  }
  if (describe_skeletons(gsm)) return 1; if (gsm_verbose) SAY("\n"); 
  if (gsm_verbose) {
    if (describe_assignments(gsm)) return 2;; SAY("\n"); 
    if (describe_positions(gsm)) return 3; SAY("\n"); 
    if (describe_blended_positions(gsm)) return 4; SAY("\n");
    if (describe_texture_coords(gsm)) return 5; SAY("\n");
    if (describe_vertices(gsm)) return 6; SAY("\n");
    if (describe_triangles(gsm)) return 7; SAY("\n");
    if (describe_triangle_groups(gsm)) return 8; SAY("\n");
    if (describe_shared_triangle_lists(gsm)) return 9; SAY("\n");
  }
  if (describe_textures(gsm)) return 10; if (gsm_verbose) SAY("\n");
  if (describe_materials(gsm)) return 11;
  if (describe_grips(gsm)) return 12; if (gsm_verbose) SAY("\n");
  if (describe_grippers(gsm)) return 13; if (gsm_verbose) SAY("\n");

  /* classify */
  if (gsm_verbose) {
    SAY("GSM type of ");
    show_gsm(gsm);
  }

  return 0;
}
