/* Milkshape .ms3d file stuff */
#ifndef MS3D_H
#define MS3D_H

#define MS3D_MAGIC "MS3D000000"
#define MS3D_VERSION 4

#define MS3D_NAME_LENGTH 32
#define MS3D_TEXTURE_LENGTH 128

#define MS3D_FLAG_SELECTED 1
#define MS3D_FLAG_HIDDEN 2
#define MS3D_FLAG_SELECTED2 4
#define MS3D_FLAG_DIRTY 8

typedef struct {
  char magic[10];
  int version;
} ms3d_header_t;

typedef struct {
  unsigned char flags; /* SELECTED | SELECTED2 | HIDDEN */
  vector_t position; /* ms3d spec uses float[3] */
  char bone_id; /* GSM uses unsigned short */
  unsigned char reference_count; /* number of triangles containing the vertex */
  unsigned short position_id; /* for eliminating dup positions in export */
  unsigned short vertex_id; /* for eliminating dup vertices in export */
} ms3d_vertex_t;

typedef struct {
  unsigned short flags; /* SELECTED | SELECTED2 | HIDDEN */
  unsigned short vertex_id[3];
  vector_t normals[3]; /* ms3d spec uses float[3][3]; not in GSM */
  float s[3]; /* U */
  float t[3]; /* V */
  unsigned char smoothing_group; /* GSM requires all one group */
  unsigned char group_id; /* triangle_group_id */
} ms3d_triangle_t;

typedef struct {
  unsigned char flags; /* SELECTED | HIDDEN */
  char name[MS3D_NAME_LENGTH];
  unsigned short num_triangles;
  unsigned short **triangle_id;
  char material_id; /* -1 => no material; GSM uses unsigned short */
} ms3d_group_t;

typedef struct {
  char name[MS3D_NAME_LENGTH];
  float ambient[4];
  float diffuse[4];
  float specular[4];
  float emissive[4];
  float shininess; /* 0.0 - 128.0 */
  float opacity; /* ms3d spec calls it transparency, this is wrong; 0.0 - 1.0 */
  char mode; /* ms3d spec says "0, 1, 2 is unused now" */
  char texture[MS3D_TEXTURE_LENGTH]; /* GSM uses char[65536] */
  char alphamap[MS3D_TEXTURE_LENGTH]; /* not in GSM */
} ms3d_material_t;

typedef struct {
  float time;
  euler_t rotation; /* ms3d spec uses float[3] */
} ms3d_keyframe_rot_t;

typedef struct {
  float time;
  vector_t position; /* ms3d spec uses float[3] */
} ms3d_keyframe_pos_t;

typedef struct {
  unsigned char flags; /* SELECTED | DIRTY */
  char name[MS3D_NAME_LENGTH]; /* GSM uses char[65536] */
  char parent_name[MS3D_NAME_LENGTH]; /* GSM uses char[65536] */
  euler_t rotation; /* ms3d spec uses float[3] */
  vector_t position; /* ms3d spec uses float[3] */
  unsigned short num_rotation_keyframes;
  unsigned short num_position_keyframes;
  ms3d_keyframe_rot_t **rotation_keyframes;
  ms3d_keyframe_pos_t **position_keyframes;
} ms3d_joint_t;

typedef struct {
  char *filename;
  char *basename;
  FILE *fd;
  FILE *out_fd;
  ms3d_header_t header;
  unsigned short num_vertices;
  unsigned short num_triangles;
  unsigned short num_groups; /* num_triangle_groups */
  unsigned short num_materials;
  float framerate;
  float current_time; /* irrelevant */
  int num_frames;
  unsigned short num_joints;
  ms3d_vertex_t **vertices;
  ms3d_triangle_t **triangles;
  ms3d_group_t **groups;
  ms3d_material_t **materials;
  ms3d_joint_t **joints;
} ms3d_model_t;

GSM_API ms3d_model_t *alloc_ms3d_model();
GSM_API void cleanup_ms3d_model(ms3d_model_t *);
GSM_API ms3d_vertex_t **alloc_ms3d_vertices(unsigned short);
GSM_API ms3d_vertex_t *alloc_ms3d_vertex();
GSM_API ms3d_triangle_t **alloc_ms3d_triangles(unsigned short);
GSM_API ms3d_triangle_t *alloc_ms3d_triangle();
GSM_API ms3d_group_t **alloc_ms3d_groups(unsigned short);
GSM_API ms3d_group_t *alloc_ms3d_group();
GSM_API void cleanup_ms3d_group(ms3d_group_t *);
GSM_API ms3d_material_t **alloc_ms3d_materials(unsigned short);
GSM_API ms3d_material_t *alloc_ms3d_material();
GSM_API ms3d_keyframe_rot_t **alloc_ms3d_rotation_keyframes(unsigned short);
GSM_API ms3d_keyframe_rot_t *alloc_ms3d_rotation_keyframe();
GSM_API ms3d_keyframe_pos_t **alloc_ms3d_position_keyframes(unsigned short);
GSM_API ms3d_keyframe_pos_t *alloc_ms3d_position_keyframe();
GSM_API ms3d_joint_t **alloc_ms3d_joints(unsigned short);
GSM_API ms3d_joint_t *alloc_ms3d_joint();
GSM_API int ms3d_truncate_string(char *, char *, int);
GSM_API int read_ms3d_file(ms3d_model_t *);
GSM_API int read_ms3d_data(ms3d_model_t *);
GSM_API int read_ms3d_header(ms3d_model_t *);
GSM_API int check_ms3d_magic(ms3d_model_t *);
GSM_API int read_ms3d_vertices(ms3d_model_t *);
GSM_API int read_ms3d_triangles(ms3d_model_t *);
GSM_API int read_ms3d_groups(ms3d_model_t *);
GSM_API int read_ms3d_materials(ms3d_model_t *);
GSM_API int read_ms3d_animation_stuff(ms3d_model_t *);
GSM_API int read_ms3d_joints(ms3d_model_t *);
GSM_API int ms3d_to_gsm(struct gsm *, ms3d_model_t *);
GSM_API int ms3d_to_gsm_process_vertices(struct gsm *, ms3d_model_t *);
GSM_API int ms3d_to_gsm_process_triangles(struct gsm *, ms3d_model_t *);
GSM_API int ms3d_to_gsm_process_triangle_groups(struct gsm *, ms3d_model_t *);
GSM_API int ms3d_to_gsm_process_materials(struct gsm *, ms3d_model_t *);
GSM_API int ms3d_to_gsm_process_joints(struct gsm *, ms3d_model_t *);
GSM_API int ms3d_to_gsm_create_grips(struct gsm *, ms3d_model_t *);
GSM_API int write_ms3d_data(ms3d_model_t *);
GSM_API int write_ms3d_header(ms3d_model_t *);
GSM_API int write_ms3d_vertices(ms3d_model_t *);
GSM_API int write_ms3d_triangles(ms3d_model_t *);
GSM_API int write_ms3d_groups(ms3d_model_t *);
GSM_API int write_ms3d_materials(ms3d_model_t *);
GSM_API int write_ms3d_animation_stuff(ms3d_model_t *);
GSM_API int write_ms3d_joints(ms3d_model_t *);
GSM_API int gsm_to_ms3d(ms3d_model_t *, struct gsm *);
GSM_API int gsm_to_ms3d_create_vertices(ms3d_model_t *, struct gsm *);
GSM_API int gsm_to_ms3d_create_triangles(ms3d_model_t *, struct gsm *);
GSM_API int gsm_to_ms3d_create_groups(ms3d_model_t *, struct gsm *);
GSM_API int gsm_to_ms3d_create_materials(ms3d_model_t *, struct gsm *);
GSM_API int gsm_to_ms3d_create_joints(ms3d_model_t *, struct gsm *);
GSM_API int gsm_to_ms3d_create_animation_stuff(ms3d_model_t *, struct gsm *);

/* transform matrix to change Milkshape co-ordinates into SWAT co-ordinates */
static matrix_t transform_ms3d_to_gsm = {
/* R0  R1  R2      R3  R4  R5      R6  R7  R8                    */
/* we'll know if this works when we get some guns */
   0,  0,  1,  0,  0,  1,  0,  0, -1,  0,  0,  0,  0,  0,  0,  1
};

/* transform matrix to change SWAT co-ordinates into Milkshape co-ordinates */
static matrix_t transform_gsm_to_ms3d = {
/* R0  R1  R2      R3  R4  R5      R6  R7  R8                    */
/* 0,  0, -1,  0,  0,  1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  1 */
/* use I for now; dunno what it should be */
   1,  0,  0,  0,  0,  1,  0,  0,  0,  0,  1,  0,  0,  0,  0,  1
};

#endif
