#include "gsmhacking.h"

struct transplant_set *alloc_transplant_set() {
  return (struct transplant_set *) alloc_one("transplant_set", sizeof(struct transplant_set));
}

void cleanup_transplant_sets(struct transplant_set *transplant_sets) {
  struct transplant_set *tmp;
  GLL_LOOP(tmp, transplant_sets) {
    TRACE2("freeing transplant_set %p file, bone and gsm\n", tmp);
    if (tmp->file_arg) free(tmp->file_arg);
    if (tmp->first_bone_arg) free(tmp->first_bone_arg);
    if (tmp->last_bone_arg) free(tmp->last_bone_arg);
    if (tmp->gsm) cleanup_gsm(tmp->gsm);
  }
  TRACE2("freeing transplant_sets at %p\n", transplant_sets);
  gll_free((struct gll *) transplant_sets);
}

int add_to_transplant_set(struct transplant_set **transplant_sets, char *file_arg, char *first_bone_arg, char *last_bone_arg, char *type) {
  struct transplant_set *last, *tmp;
  int first_bone_id, last_bone_id;
  char *bogus;
  unsigned short len;
 
  tmp = alloc_transplant_set();
  if (! tmp) {
    ERROR("out of memory for transplant_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  tmp->file_arg = alloc_name(file_arg, &len);
  if (! len) {
    ERROR("out of memory for transplant_set->file_arg");
    return 2;
  }
  TRACE("tmp->file_arg: %p\n", tmp->file_arg);

  /* get first bone to process */
  first_bone_id = (int) strtol(first_bone_arg, &bogus, 0);
  if (*bogus) {
    /*
      bone argument isn't a number, assume bone name
      we can't deal with it until the GSM is loaded, so store it for later 
      processing
    */
    tmp->first_bone_arg = alloc_name(first_bone_arg, &len);
    if (! len) {
      ERROR("out of memory for transplant_set->first_bone_arg");
      return 3;
    }
    TRACE("tmp->first_bone_arg: %p\n", tmp->first_bone_arg);
  }
  else {
    /* could be a direct bone ID */
    if (first_bone_id < 0) {
      ERROR("negative first bone ID\n");
      return 4;
    }
    if (first_bone_id > 0xffff) {
      ERROR("first bone ID out of range\n");
      return 5;
    }
    tmp->first_bone_id = (unsigned short) first_bone_id;
    TRACE("tmp->first_bone_id: %u\n", tmp->first_bone_id);
  }

  /* get last bone; will do all bones from first through last */
  last_bone_id = (int) strtol(last_bone_arg, &bogus, 0);
  if (*bogus) {
    tmp->last_bone_arg = alloc_name(last_bone_arg, &len);
    if (! len) {
      ERROR("out of memory for transplant_set->last_bone_arg");
      return 6;
    }
    TRACE("tmp->last_bone_arg: %p\n", tmp->last_bone_arg);
  }
  else {
    /* could be a direct bone ID */
    if (last_bone_id > 0xffff) {
      ERROR("last bone ID out of range\n");
      return 8;
    }
    if (last_bone_id < first_bone_id) {
      ERROR("last bone ID < first bone ID!\n");
      return 9;
    }
    tmp->last_bone_id = (unsigned short) last_bone_id;
    TRACE("tmp->last_bone_id: %u\n", tmp->last_bone_id);
  }

  /* type of transplant: normal or magic */
  if (str_equiv(type, "normal")) {
    tmp->type = TRANSPLANT_TYPE_NORMAL;
    TRACE("normal transplant\n");
  }
  else if (str_equiv(type, "magic")) {
    tmp->type = TRANSPLANT_TYPE_MAGIC;
    TRACE("magic transplant\n");
  }
  else {
    ERROR("bogus transplant type; try \"normal\" or \"magic\"\n");
    return 9;
  }
  TRACE("tmp->type: %u\n", tmp->type);

  last  = (struct transplant_set *) gll_last((struct gll *) *transplant_sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *transplant_sets = tmp;
  /*gll_debug((struct gll *) *transplant_sets);*/

  return 0;
}

int transplant_bones(struct gsm *gsm) {
  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (gsm->is_animation) return anim_transplant_bones(gsm);

  /* XXX */
  error_dbg("mesh transplants not yet supported, use --transplant for animations only\n");
  return -1;
}

int anim_transplant_bones(struct gsm *gsa) {
  struct transplant_set *transplant_set, *transplant_sets;
  int errors, ret;

  if (! gsa) {
    ERROR("no GSA!\n");
    return 1;
  }

  transplant_sets = gsa->option->transplant_sets;
  if (! transplant_sets) {
    ERROR("no transplant_sets\n");
    return 2;
  }

  errors = 0;
  GLL_LOOP(transplant_set, transplant_sets) {
    ret = anim_prep_bone_transplant(gsa, transplant_set);
    if (ret) {
      ERROR("anim_prep_bone_transplant() returned %d\n", ret);
      errors++;
    }
  }

  if (errors) {
    ERROR("%d error%s occurred during surgery\n", errors, (errors == 1) ? "" : "s");
    errors += 2;
  }
  return errors;
}

int anim_prep_bone_transplant(struct gsm *gsa, struct transplant_set *transplant_set) {
  struct gsm *donor;
  int ret;
  unsigned short len;

  if (! gsa) {
    ERROR("no GSA!\n");
    return 1;
  }
  if (! transplant_set) {
    ERROR("no transplant_set\n");
    return 2;
  }
  if (! transplant_set->file_arg) {
    ERROR("no GSA argument in transplant_set\n");
    return 3;
  }

  donor = alloc_gsm();
  if (! donor) {
    ERROR("out of memory for donor GSA");
    return 4;
  }

  donor->filename = alloc_name(transplant_set->file_arg, &len);
  if (! len) {
    ERROR("out of memory for donor filename %s", transplant_set->file_arg);
    return 5;
  }

  /* read in the donor */
  if (ret = read_gsa_file(donor)) {
    if (ret > 0) ERROR("%s isn't a gsa file\n", donor->filename);
    else ERROR("can't open %s\n", donor->filename, ret);
    cleanup_gsm(donor);
    return 6;
  }
  if (parse_gsa(donor)) {
    ERROR("can't parse animation\n");
    cleanup_gsm(donor);
    return 7;
  }
  classify_gsa(donor);

  /* we NEED a new-style animation */
  if (gsa->is_animation && gsa->version >> 8 == 4 && gsa->version % 16 < 2) {
    SAY("%s is in the old format without event support; optimising!\n", gsa->basename);
    if (optimise(gsa)) {
      ERROR("optimisation failed, can't continue\n");
      cleanup_gsm(donor);
      return 8;
    }
  }
  if (donor->is_animation && donor->version >> 8 == 4 && donor->version % 16 < 2) {
    SAY("%s is in the old format without event support; optimising!\n", donor->basename);
    if (optimise(donor)) {
      ERROR("optimisation failed, can't continue\n");
      cleanup_gsm(donor);
      return 8;
    }
  }

  /* can only transplant from an animation of the same class */
  if (donor->type % GSM_TYPE_ANIMATION_LIMITED_BONES != gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES) {
    ERROR("type mismatch: %s is type %u and %s is type %u\n", gsa->basename, gsa->type % GSM_TYPE_ANIMATION_LIMITED_BONES, donor->basename, donor->type % GSM_TYPE_ANIMATION_LIMITED_BONES);
    cleanup_gsm(donor);
    return 9;
  }

  /* NOW we can parse the bone arguments */
  if (transplant_set->first_bone_arg) {
    ret = get_bone_by_name(donor, transplant_set->first_bone_arg);
    if (ret < 0) {
      ERROR("can't transplant a non-existent bone (first)!\n");
      cleanup_gsm(donor);
      return 10;
    }
    if (ret >= gsa->skeletons[0]->num_bones) {
      ERROR("returned first bone ID out of target's range: this is impossible!\n");
      cleanup_gsm(donor);
      return 11;
    }
    transplant_set->first_bone_id = (unsigned short) ret;
  }

  if (transplant_set->last_bone_arg) {
    ret = get_bone_by_name(donor, transplant_set->last_bone_arg);
    if (ret < 0) {
      ERROR("can't transplant a non-existent bone (last)!\n");
      cleanup_gsm(donor);
      return 12;
    }
    if (ret >= gsa->skeletons[0]->num_bones) {
      ERROR("returned last bone ID out of target's range: this is impossible!\n");
      cleanup_gsm(donor);
      return 13;
    }
    transplant_set->last_bone_id = (unsigned short) ret;
    if (transplant_set->last_bone_id < transplant_set->first_bone_id) {
      ERROR("last bone ID < first bone ID\n");
      return 14;
    }
  }

  /* we need a new function for all this crap */
  return anim_perform_bone_transplant(gsa, donor, transplant_set);
}

/* helper function; don't call directly */
int anim_perform_bone_transplant(struct gsm *patient, struct gsm *donor, struct transplant_set *transplant_set) {
  unsigned short i, j, first_bone_id, last_bone_id, patient_orientation_id, donor_orientation_id;

  if (! patient || ! donor || ! transplant_set) {
    ERROR("operating theatre is lacking a few key personalities\n");
    error_dbg("read the source if you want to know wtf that is supposed to mean\n");
    error_dbg("there are only so many times a cat is willing to perform the SAME\n");
    error_dbg("sanity checks on the SAME data\n");
    if (donor) cleanup_gsm(donor);
    return 15;
  }

  /* the transplant set bone_id is the skeleton's bone ID */
  first_bone_id = transplant_set->first_bone_id;
  if (first_bone_id >= patient->skeletons[0]->num_bones) {
    ERROR("first bone ID %u out of target's range\n", first_bone_id);
    cleanup_gsm(donor);
    return 16;
  }
  if (first_bone_id >= donor->skeletons[0]->num_bones) {
    ERROR("first bone ID %u out of sources's range: this should be impossible!\n", first_bone_id);
    cleanup_gsm(donor);
    return 17;
  }
  /*
    but if not all bones are animated, this does not match the orientation 
    we are looking ofr
  */
  if (! patient->skeletons[0]->bones[first_bone_id]->animated) {
    /* FIXME: may be a problem depending on the animation type */
    WARN("first bone %u is not animated in target skeleton\n");
    /*return 18;*/
  }
  if (! donor->skeletons[0]->bones[first_bone_id]->animated) {
    ERROR("first bone %u is not animated in source skeleton\n");
    cleanup_gsm(donor);
    return 19;
  }

  last_bone_id = transplant_set->last_bone_id;
  if (last_bone_id >= patient->skeletons[0]->num_bones) {
    ERROR("last bone ID %u out of target's range\n", last_bone_id);
    cleanup_gsm(donor);
    return 22;
  }
  if (last_bone_id >= donor->skeletons[0]->num_bones) {
    ERROR("last bone ID %u out of sources's range: this should be impossible!\n", last_bone_id);
    cleanup_gsm(donor);
    return 23;
  }
  /*
    but if not all bones are animated, this does not match the orientation 
    we are looking ofr
  */
  if (! patient->skeletons[0]->bones[last_bone_id]->animated) {
    /* FIXME: may be a problem depending on the animation type */
    WARN("last bone %u is not animated in target skeleton\n");
    /*return 24;*/
  }
  if (! donor->skeletons[0]->bones[last_bone_id]->animated) {
    ERROR("last bone %u is not animated in source skeleton\n");
    cleanup_gsm(donor);
    return 25;
  }


  for (i = first_bone_id; i <= last_bone_id; i++) {
    SAY("transplanting bone %u \"%s\" from %s to %s\n", i, get_bone_by_id(patient, i), donor->basename, patient->basename);
    /* this bone may not have been animated */
    INFO("marking bone \"%s\" as animated\n", get_bone_by_id(patient, i));
    patient->skeletons[0]->bones[i]->animated = 1;

    /* set position in frame */
    if (patient->num_frames != donor->num_frames) {
      ERROR("magic transplant not ready!\n");
      cleanup_gsm(donor);
      return -1;
    }

    for (j = 0; j < patient->num_frames; j++) {
      INFO("fixing bone %u position in frame %u\n", i, j);
      if (i == first_bone_id) {
        if (anim_transplant_bone(patient, patient->skeletons[0]->bones[i], donor->skeletons[0]->bones[i], j)) {
          ERROR("failed to set bone %u's position in frame %u\n", i, j);
          cleanup_gsm(donor);
          return -2;
        }
      }
      else {
        /* cheat */
        copy_quaternion(patient->skeletons[0]->bones[i]->orientations[j], donor->skeletons[0]->bones[i]->orientations[j]);
        copy_matrix(patient->skeletons[0]->bones[i]->locals[j], donor->skeletons[0]->bones[i]->locals[j]);
        copy_matrix(patient->skeletons[0]->bones[i]->worlds[j], donor->skeletons[0]->bones[i]->worlds[j]);
      }
      if (calculate_world_for_all_children(patient, i, j)) {
        ERROR("failed to recalculate child positions\n");
        return -3;
      }
    }
  }

  cleanup_gsm(donor);

  return 0;
}

/*
  calculation angle of this bone relative to parent in donor and rotate 
  the same bone in the patient to match
*/
int anim_transplant_bone(struct gsm *gsa, struct bone *patient, struct bone *donor, unsigned short num) {
  matrix_t m, *n;
  struct bone *parent;
      
  if (! donor) {
    ERROR("no source bone\n");
    return 1;
  }
  if (! patient) {
    ERROR("no destination bone\n");
    return 2;
  }

  if (! donor->worlds || ! donor->worlds[num])  {
    ERROR("no source world rotation\n");
    return 3;
  }
  if (! patient->worlds || ! patient->worlds[num])  {
    ERROR("no destination world rotation\n");
    return 4;
  }

  /* Mparent.Mbone' = Mworld' => Mbone' = Mparent_transpose.Mworld' */
  if (patient->links_to == BONE_FIRST_LINK) {
    /* root bone - desired rotation is target world */
    copy_matrix(patient->locals[num], donor->worlds[num]);
    copy_matrix(patient->worlds[num], donor->worlds[num]);
  }
  else {
    /* transpose parent's world */
    if (! gsa) {
      ERROR("oops, where's the patient GSA?\n");
      return 5;
    }
    parent = gsa->skeletons[0]->bones[patient->links_to];
    if (! parent) {
      ERROR("null parent bone\n");
      return 6;
    }
    copy_matrix(&m, parent->worlds[num]);
    /* you know transpose_matrix() really can't fail */
    if (transpose_matrix(&m)) {
      ERROR("can't get transpose of parent world rotation matrix\n");
      return 7;
    }

    /* multiply matrices */
    n = multiply_matrices(&m, donor->worlds[num]);
    if (! n) {
      ERROR("can't multiply matrices\n");
      return 8;
    }
    copy_matrix(patient->locals[num], n);
    free(n);

    /* explicitly calculate world matrix */
    n = multiply_matrices(parent->worlds[num], patient->locals[num]);
    if (! n) {
      ERROR("can't calculate new world explicitly\n");
      return 9;
    }
    copy_matrix(patient->worlds[num], n);
    free(n);
  }

  /* convert resulting matrix to a quaternion */
  if (matrix_to_quaternion(patient->orientations[num], patient->locals[num])) {
    ERROR("can't convert local rotation matrix to quaternion\n");
    return 10;
  }

  /* test */
  if (quaternion_to_matrix(&m, patient->orientations[num])) {
    ERROR("fuck\n");
    return 11;
  }

  n = patient->locals[num];
  SAY("  Local   | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", (*n)[R0], (*n)[R3], (*n)[R6], (*n)[TX]);
  SAY("  matrix: | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", (*n)[R1], (*n)[R4], (*n)[R7], (*n)[TY]);
  SAY("          | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", (*n)[R2], (*n)[R5], (*n)[R8], (*n)[TZ]);

  SAY("\n");

  n = patient->worlds[num];
  SAY("  World   | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", (*n)[R0], (*n)[R3], (*n)[R6], (*n)[TX]);
  SAY("  matrix: | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", (*n)[R1], (*n)[R4], (*n)[R7], (*n)[TY]);
  SAY("          | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", (*n)[R2], (*n)[R5], (*n)[R8], (*n)[TZ]);
  
  SAY("\n");

  n = donor->worlds[num];
  SAY("  Target  | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", (*n)[R0], (*n)[R3], (*n)[R6], (*n)[TX]);
  SAY("  matrix: | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", (*n)[R1], (*n)[R4], (*n)[R7], (*n)[TY]);
  SAY("          | % 0.4f, % 0.4f, % 0.4f, % 0.4f |\n", (*n)[R2], (*n)[R5], (*n)[R8], (*n)[TZ]);
 
  SAY("\n");

  return 0;
}

struct copy_material_set *alloc_copy_material_set() {
  return (struct copy_material_set *) alloc_one("copy_material_set", sizeof(struct copy_material_set));
}

void cleanup_copy_material_set(struct copy_material_set *copy_material_sets) {
  struct copy_material_set *tmp;
  GLL_LOOP(tmp, copy_material_sets) {
    TRACE2("freeing copy_material_set %p file and material\n", tmp);
    if (tmp->file_arg) free(tmp->file_arg);
    if (tmp->material_arg) free(tmp->material_arg);
  }
  TRACE2("freeing copy_material_sets at %p\n", copy_material_sets);
  gll_free((struct gll *) copy_material_sets);
}

int add_to_copy_material_sets(struct copy_material_set **copy_material_sets, char *file_arg, char *material_arg, char *name_arg) {
  struct copy_material_set *last, *tmp;
  int material_id;
  char *bogus;
  unsigned short len;
 
  tmp = alloc_copy_material_set();
  if (! tmp) {
    ERROR("out of memory for copy_material_set");
    return 1;
  }
  TRACE("tmp: %p\n", tmp);

  tmp->file_arg = alloc_name(file_arg, &len);
  if (! len) {
    ERROR("out of memory for copy_material_set->file_arg");
    /* XXX: cleanup! */
    return 2;
  }
  TRACE("tmp->file_arg: %p\n", tmp->file_arg);

  /* get material */
  material_id = (int) strtol(material_arg, &bogus, 0);
  if (*bogus) {
    /*
      bone argument isn't a number, assume material name
      we can't deal with it until the GSM is loaded, so store it for later 
      processing
    */
    tmp->material_arg = alloc_name(material_arg, &len);
    if (! len) {
      ERROR("out of memory for copy_material_set->material_arg");
      return 3;
    }
    TRACE("tmp->material_arg: %p\n", tmp->material_arg);
  }
  else {
    /* could be a direct material ID */
    if (material_id < 0) {
      ERROR("negative material ID\n");
      return 4;
    }
    if (material_id > 0xffff) {
      ERROR("material ID out of range\n");
      return 5;
    }
    tmp->material_id = (unsigned short) material_id;
    TRACE("tmp->material_id: %u\n", tmp->material_id);
  }

  /* set new name */
  if (*name_arg) {
    tmp->name_arg = alloc_name(name_arg, &len);
    if (! len) {
      ERROR("out of memory for new material name");
      return 6;
    }
    TRACE("tmp->name_arg: \"%s\"\n", tmp->name_arg);
  }
  else TRACE("tmp->name_arg: \"\" - will keep same material name in destination\n");

  last  = (struct copy_material_set *) gll_last((struct gll *) *copy_material_sets);
  TRACE("last: %p\n", last);
  /* this is the first one */
  if (last) last->next = tmp;
  else *copy_material_sets = tmp;
  /*gll_debug((struct gll *) *copy_material_sets);*/

  return 0;
}

/*
  copy materials (and positions/vertices/triangles/triangle groups) from 
  another GSM

    here's what we'll do

    load in GSM
    find material by name or ID
    get all triangle groups with that material
    find all triangles in those groups
    find all vertices with those triangles
    find all positions and texture_coords in those vertices
    add positions/texture_coords/vertices/triangles/triangle_groups to GSM
    set material name
  */
int copy_materials(struct gsm *gsm) {
  struct copy_material_set *copy_material_set, *copy_material_sets;
  int errors, ret;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  if (gsm->is_animation) {
    ERROR("animations don't have materials\n");
    return -1;
  }

  copy_material_sets = gsm->option->copy_material_sets;
  if (! copy_material_sets) {
    ERROR("no copy_material_sets\n");
    return 2;
  }

  errors = 0;
  GLL_LOOP(copy_material_set, copy_material_sets) {
    ret = copy_material(gsm, copy_material_set);
    if (ret) {
      ERROR("copy_material() returned %d\n", ret);
      errors++;
    }
  }

  return errors;
}

int copy_material(struct gsm *gsm, struct copy_material_set *set) {
  struct gsm *src;
  struct material *material, *new_material;
  int ret;
  unsigned short len, num_triangles, num_triangle_groups;
  mapping_t *vertex_mapping, *position_mapping, *texture_coord_mapping;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (! set) {
    ERROR("no copy_material_set\n");
    return 2;
  }
  if (! set->file_arg) {
    ERROR("no GSM argument in copy_material_set\n");
    return 3;
  }

  src = alloc_gsm();
  if (! src) {
    ERROR("out of memory for source GSM");
    return 4;
  }

  src->filename = alloc_name(set->file_arg, &len);
  if (! len) {
    ERROR("out of memory for source filename %s\n", set->file_arg);
    return 5;
  }

  /* read in the source */
  DEBUG("reading source GSM\n");
  if (ret = read_gsm_file(src)) {
    if (ret > 0) ERROR("%s isn't a gsm file\n", src->filename);
    else ERROR("can't open %s\n", src->filename, ret);
    cleanup_gsm(src);
    return 6;
  }
  DEBUG("reading source GSM\n");
  if (parse_gsm(src)) {
    ERROR("can't parse mesh\n");
    cleanup_gsm(src);
    return 7;
  }
  classify_gsm(src);

  if (src->is_animation) {
    ERROR("can't copy materials from an animation\n");
    cleanup_gsm(src);
    return 8;
  }

  /* now parse out arguments from copy_material_set */
  if (set->material_arg) {
    ret = get_material_by_name(src, set->material_arg);
    if (ret < 0) {
      ERROR("can't transplant a non-existent material!\n");
      cleanup_gsm(src);
      return 9;
    }
    set->material_id = (unsigned short) ret;
  }

  /* source material */
  material = src->materials[set->material_id];
  if (! material) {
    ERROR("null material - hmm!\n");
    cleanup_gsm(src);
    return 10;
  }

  /* set new material name if we want to keep the original */
  if (! set->name_arg) {
    set->name_arg = alloc_name(material->name, &len);
    if (! len) {
      ERROR("out of memory for new material name");
      cleanup_gsm(src);
      return 11;
    }
  }

  /*
    we increase ret each time so we can have a clean goto rather than typing
    ERROR(), cleanup_gsm(), return over and over again
  */
  ret = 12; if (copy_material_create_material(gsm, src, set)) goto oops;
  ret++; if (copy_material_create_triangles(gsm, src, set, &num_triangles, &num_triangle_groups)) goto oops;
  ret++; if (copy_material_create_vertices(gsm, src, num_triangles, num_triangle_groups, &vertex_mapping)) goto oops;
  len = vertex_mapping->size;
  cleanup_mapping(vertex_mapping);
  ret++; if (copy_material_create_shared_triangle_lists(gsm, src, num_triangles, len)) goto oops;
  ret++; if (copy_material_create_texture_coords(gsm, src, len, &texture_coord_mapping)) goto oops;
  ret++; if (copy_material_create_positions(gsm, src, len, &position_mapping)) goto oops;

  cleanup_gsm(src);
  return 0;

oops:
  ERROR("can't perform material copying operation\n");
  cleanup_gsm(src);
  return ret;
}

/* make the new material and any maps/textures */
int copy_material_create_material(struct gsm *dest, struct gsm *src, struct copy_material_set *set) {
  struct material *material, *new_material, **new_materials;
  struct texture *texture, *new_texture, **new_textures;
  struct map *map, *new_map;
  unsigned short i, num_textures, len, first_texture;

  if (! dest || ! src || ! set) {
    ERROR("missing arguments\n");
    return 1;
  }

  /* the original material */
  material = src->materials[set->material_id];
  DEBUG("source material \"%s\"\n", material->name);

  /* make new materials */
  DEBUG("making 1 new material\n");
  new_materials = alloc_more_materials(1, dest->num_materials, dest->materials);
  if (! new_materials) {
    ERROR("out of memory for new materials");
    return 2;
  }
  dest->materials = new_materials;
  /* the new material */
  TRACE("adding new material\n");
  new_material = dest->materials[dest->num_materials++];
  /* copy original material data - we'll overwrite the new stuff later */
  memmove(new_material, material, sizeof(struct material));
  new_material->offset = 0x0fc0;

  /* set name */
  new_material->name = alloc_name(set->name_arg, &len);
  if (! len) {
    ERROR("out of memory for new material name");
    return 3;
  }

  /* make new maps */
  DEBUG("making new maps\n");
  if (material->num_maps) {
    new_material->maps = alloc_maps(material->num_maps);
    if (! new_material->maps) {
      ERROR("out of memory for new maps");
      /* will be freed when cleanup_gsm() is called on src */
      new_material->num_maps = 0;
      return 4;
    }
    new_material->num_maps = material->num_maps;

    /* count new textures */
    num_textures = material->num_maps;

    /* allocate new textures */
    DEBUG("making %u new texture%s\n", num_textures, (num_textures == 1) ? "" : "s");
    new_textures = alloc_more_textures(num_textures, dest->num_textures, dest->textures);
    if (! new_textures) {
      ERROR("out of memory for new textures");
      return 5;
    }
    dest->textures = new_textures;
    first_texture = dest->num_textures;
    dest->num_textures += num_textures;
    /* add new ones */
    for (i = 0; i < material->num_maps; i++) {
      map = material->maps[i];
      texture = src->textures[map->texture_id];
      new_texture = new_textures[first_texture + i];
      /* set new map */
      TRACE("setting new map %u\n", i);
      new_map = new_material->maps[i];
      new_map->map_type = map->map_type;
      new_map->texture_id = first_texture + i;
      /* set new texture */
      TRACE("setting new texture %u \"%s\"\n", i, texture->basename);
      new_texture->offset = 0x0fc0;
      new_texture->filename = alloc_name(texture->filename, &new_texture->filename_length);
      if (! new_texture->filename_length) {
        ERROR("out of memory for new texture filename");
        cleanup_textures(dest->num_textures + num_textures, new_textures);
        return 6;
      }
      set_basename(new_texture->filename, new_texture->filename_length, &new_texture->basename, &new_texture->basename_length);
    }
  }

  return 0;
}

/* copy triangle groups mapped to the material we're copying */
int copy_material_create_triangles(struct gsm *dest, struct gsm *src, struct copy_material_set *set, unsigned short *num_triangles, unsigned short *num_triangle_groups) {
  struct material *material;
  struct triangle_group *triangle_group, *new_triangle_group, **new_triangle_groups;
  struct triangle *triangle, *new_triangle, **new_triangles;
  unsigned short i, j, n, count, first_triangle, first_triangle_group;

  if (! dest || ! src || ! set) {
    ERROR("missing arguments\n");
    return 1;
  }

  /* the original material */
  material = src->materials[set->material_id];
  DEBUG("source material \"%s\"\n", material->name);

  /* count triangle groups and triangles mapped to this material */
  *num_triangle_groups = *num_triangles = 0;
  for (i = 0; i < src->num_triangle_groups; i++) {
    triangle_group = src->triangle_groups[i];
    if (triangle_group->material_id != set->material_id) continue;
    ++*num_triangle_groups;
    *num_triangles += triangle_group->num_triangles;
  }

  /* make new triangles */
  DEBUG("making %u new triangle%s\n", *num_triangles, (*num_triangles == 1) ? "" : "s");
  new_triangles = alloc_more_triangles(*num_triangles, dest->num_triangles, dest->triangles);
  if (! new_triangles) {
    ERROR("out of memory for new triangles");
    return 2;
  }
  dest->triangles = new_triangles;
  first_triangle = dest->num_triangles;
  dest->num_triangles += *num_triangles;

  /* make new triangle groups */
  DEBUG("making %u new triangle group%s\n", *num_triangle_groups, (*num_triangle_groups == 1) ? "" : "s");
  new_triangle_groups = alloc_more_triangle_groups(*num_triangle_groups, dest->num_triangle_groups, dest->triangle_groups);
  if (! new_triangle_groups) {
    ERROR("out of memory for new triangle groups");
    return 3;
  }
  dest->triangle_groups = new_triangle_groups;
  first_triangle_group = dest->num_triangle_groups;
  dest->num_triangle_groups += *num_triangle_groups;

  /* set number of triangles and material; start counting new triangles */
  count = n = 0;
  for (i = 0; i < src->num_triangle_groups; i++) {
    triangle_group = src->triangle_groups[i];
    if (triangle_group->material_id != set->material_id) continue;
    TRACE("setting new triangle group %u from source group %u\n", n, i);
    new_triangle_group = new_triangle_groups[first_triangle_group + n];
    /* new triangles bolted on to end; keep track of number of new ones */
    new_triangle_group->offset = 0x0fc0;
    new_triangle_group->first_triangle = first_triangle + count;
    new_triangle_group->num_triangles = triangle_group->num_triangles;
    /* new material was bolted on to end */
    new_triangle_group->material_id = dest->num_materials - 1;
    /* set triangle vertices for later processing */
    for (j = 0; j < triangle_group->num_triangles; j++) {
      triangle = src->triangles[triangle_group->first_triangle + j];
      new_triangle = new_triangles[first_triangle + count + j];
      new_triangle->offset = 0x0fc0;
      new_triangle->vertex1 = triangle->vertex1;
      new_triangle->vertex2 = triangle->vertex2;
      new_triangle->vertex3 = triangle->vertex3;
      TRACE("copying triangle %u as %u (%u, %u, %u)\n", triangle_group->first_triangle + j, first_triangle + j, triangle->vertex1, triangle->vertex2, triangle->vertex3);
    }
    /* next new triangle group */
    n++;
    count += new_triangle_group->num_triangles;
  }

  return 0;
}

/* find unique vertices/texture_coords/positions from new triangle groups */
int copy_material_create_vertices(struct gsm *dest, struct gsm *src, unsigned short num_triangles, unsigned short num_triangle_groups, mapping_t **vertex_mapping) {
  struct position **new_positions;
  struct triangle *triangle;
  struct vertex *vertex, *new_vertex, **new_vertices;
  unsigned short i, j, m, n, num_vertices, first_triangle, first_vertex;
  int tmp;

  if (! dest || ! src) {
    ERROR("missing arguments\n");
    return 1;
  }
  /* unlikely to happen */
  if (num_triangles > 0xffff / 3) {
    ERROR("too many triangles\n");
    return 2;
  }
  num_vertices = num_triangles * 3;
  if (num_vertices == 0xffff) {
    error_dbg("the gsmhacking code assumes no mesh will ever have %u vertices\n", 0xffff);
    error_dbg("if the mesh really does have that many, and if it doesn't crash the game\n");
    error_dbg("send your GSM to furrycat and tell him you got this message\n");
    return 2;
  }

  /* first (as yet) unused triangle in new_triangles */
  first_triangle = dest->num_triangles - num_triangles;
  /* first unused vertex in new_vertices (new ones not made yet!) */
  first_vertex = dest->num_vertices;

  DEBUG("creating vertex mapping of size %u\n", num_vertices);
  *vertex_mapping = alloc_mapping(num_vertices);
  if (! *vertex_mapping) {
    ERROR("out of memory for vertex mapping");
    return 3;
  }
  /* set default values in hash */
  initialise_mapping(*vertex_mapping);

  /* fill up the "hash" */
  n = 0;
  for (i = 0; i < num_triangles; i++) {
    triangle = dest->triangles[first_triangle + i];
    /* first vertex */
    tmp = get_key_for_mapping(*vertex_mapping, triangle->vertex1);
    if (tmp < -1) {
      ERROR("can't get mapping for vertex one in new triangle %u\n", i);
      cleanup_mapping(*vertex_mapping);
      return 4;
    }
    if (tmp == -1 || tmp == 0xffff) {
      if (set_value_for_mapping(*vertex_mapping, n, triangle->vertex1)) {
        ERROR("can't set mapping for vertex one %u to %u in new triangle %u\n", triangle->vertex1, n, i);
        cleanup_mapping(*vertex_mapping);
        return 5;
      }
      n++;
    }
    /* second vertex */
    tmp = get_key_for_mapping(*vertex_mapping, triangle->vertex2);
    if (tmp < -1) {
      ERROR("can't get mapping for vertex two in new triangle %u\n", i);
      cleanup_mapping(*vertex_mapping);
      return 6;
    }
    if (tmp == -1 || tmp == 0xffff) {
      if (set_value_for_mapping(*vertex_mapping, n, triangle->vertex2)) {
        ERROR("can't set mapping for vertex two %u to %u in new triangle %u\n", triangle->vertex2, n, i);
        cleanup_mapping(*vertex_mapping);
        return 7;
      }
      n++;
    }
    /* third vertex */
    tmp = get_key_for_mapping(*vertex_mapping, triangle->vertex3);
    if (tmp < -1) {
      ERROR("can't get mapping for vertex three in new triangle %u\n", i);
      cleanup_mapping(*vertex_mapping);
      return 8;
    }
    if (tmp == -1 || tmp == 0xffff) {
      if (set_value_for_mapping(*vertex_mapping, n, triangle->vertex3)) {
        ERROR("can't set mapping for vertex three %u to %u in new triangle %u\n", triangle->vertex3, n, i);
        cleanup_mapping(*vertex_mapping);
        return 9;
      }
      n++;
    }
  }
  
  /* eliminate unused entries in the mapping */
  for (i = (*vertex_mapping)->size - 1; i > 0; i--) {
    tmp = get_value_for_mapping(*vertex_mapping, i);
    if (tmp < -1) {
      ERROR("oh dear, vertex mapping is bollocksed\n");
      cleanup_mapping(*vertex_mapping);
      return 10;
    }
    if (tmp > -1 && tmp < 0xffff) break;
    DEBUG("redundant entry %u of %u in vertex mapping\n", i, (*vertex_mapping)->size);
    pop_mapping(*vertex_mapping, i);
  }
  /* and sort them */
  sort_mapping_by_value(*vertex_mapping, SORT_MAPPING_NORMALISED);
  debug_mapping(*vertex_mapping, "vertex");

  /* now we have the vertices in each triangle */
  num_vertices = (*vertex_mapping)->size;
  /*SAY("first vertex is %u\n", first_vertex);*/
  for (i = 0; i < num_triangles; i++) {
    triangle = dest->triangles[first_triangle + i];
    /* let's assume - just this once - that the vertex mapping is OK */
    triangle->vertex1 = (unsigned short) get_key_for_mapping(*vertex_mapping, triangle->vertex1) + first_vertex;
    triangle->vertex2 = (unsigned short) get_key_for_mapping(*vertex_mapping, triangle->vertex2) + first_vertex;
    triangle->vertex3 = (unsigned short) get_key_for_mapping(*vertex_mapping, triangle->vertex3) + first_vertex;
    TRACE("setting triangle %u (new triangle %u) vertices (%u, %u, %u)\n", first_triangle + i, i, triangle->vertex1, triangle->vertex2, triangle->vertex3);
  }

  /* (finally) make new vertices */
  DEBUG("making %u new vertice%s\n", num_vertices, (num_vertices == 1) ? "" : "s");
  new_vertices = alloc_more_vertices(num_vertices, dest->num_vertices, dest->vertices);
  if (! new_vertices) {
    ERROR("out of memory for new vertices");
    cleanup_mapping(*vertex_mapping);
    return 16;
  }
  dest->vertices = new_vertices;
  first_vertex = dest->num_vertices;
  dest->num_vertices += num_vertices;
  /* add new ones from source */
  for (i = 0; i < num_vertices; i++) {
    new_vertex = new_vertices[i + first_vertex];
    tmp = get_value_for_mapping(*vertex_mapping, i);
    if (tmp < 0) {
      ERROR("can't get original vertex from map index %u\n", i);
      cleanup_mapping(*vertex_mapping);
      return 17;
    }
    vertex = src->vertices[tmp];
    new_vertex->offset = 0x0fc0;
    /* texture co-ordinate can be mapped later */
    new_vertex->texture_coord_id = vertex->texture_coord_id;
    /* position can be mapped later */
    new_vertex->position_id = vertex->position_id;
    /* XXX: shared triangle list */
    new_vertex->shared_triangle_list_id = vertex->shared_triangle_list_id;
    TRACE("copying vertex %u as %u; position %u, texture_coord %u\n", (unsigned short) tmp, i, vertex->position_id, vertex->texture_coord_id);
  }

  return 0;
}

/* just make some bogus lists */
int copy_material_create_shared_triangle_lists(struct gsm *dest, struct gsm *src, unsigned short num_triangles, unsigned short num_vertices) {
  struct shared_triangle_list *new_shared_triangle_list;
  struct triangle *triangle;
  struct vertex *vertex;
  unsigned short i, first_triangle, first_vertex, first_shared_triangle_list;

  if (! dest || ! src) {
    ERROR("missing arguments\n");
    return 1;
  }

  if (num_vertices == 0xffff) {
    error_dbg("the gsmhacking code assumes no mesh will ever have %u vertices\n", 0xffff);
    error_dbg("if the mesh really does have that many, and if it doesn't crash the game\n");
    error_dbg("send your GSM to furrycat and tell him you got this message\n");
    return 2;
  }

  /* first (as yet) unused triangle in new_triangles */
  first_triangle = dest->num_triangles - num_triangles;
  /* first unused vertex in new_vertices */
  first_vertex = dest->num_vertices - num_vertices;

  DEBUG("creating cheap and cheerful shared triangle lists\n");
  /* make one list for each triagle */
  dest->shared_triangle_lists = alloc_more_shared_triangle_lists(num_triangles, dest->num_shared_triangle_lists, dest->shared_triangle_lists);
  first_shared_triangle_list = dest->num_shared_triangle_lists;
  if (! dest->shared_triangle_lists) {
    ERROR("out of memory for shared triangle lists");
    return 3;
  }
  dest->num_shared_triangle_lists += num_triangles;

  for (i = 0; i < num_triangles; i++) {
    triangle = dest->triangles[i];
    new_shared_triangle_list = dest->shared_triangle_lists[first_shared_triangle_list + i];
    /* create the new bogus list */
    new_shared_triangle_list->num_triangles = 1;
    new_shared_triangle_list->triangles = (unsigned short **) alloc_lots(new_shared_triangle_list->num_triangles, "shared_triangle", sizeof(unsigned short *), (void *(*)()) alloc_short);
    if (! new_shared_triangle_list->triangles) {
      ERROR("out of memory for list");
      return 4;
    }
    *new_shared_triangle_list->triangles[0] = first_triangle + i;
    /* set vertices in the list */
    vertex = dest->vertices[triangle->vertex1];
    vertex->shared_triangle_list_id = first_shared_triangle_list + i;
    vertex = dest->vertices[triangle->vertex2];
    vertex->shared_triangle_list_id = first_shared_triangle_list + i;
    vertex = dest->vertices[triangle->vertex3];
    vertex->shared_triangle_list_id = first_shared_triangle_list + i;
  }
  dest->shared_triangle_list_length += num_triangles * 2;

  return 0;
}

int copy_material_create_texture_coords(struct gsm *dest, struct gsm *src, unsigned short num_vertices, mapping_t **texture_coord_mapping) {
  struct texture_coord *texture_coord, *new_texture_coord, **new_texture_coords;
  struct vertex *vertex;
  unsigned short i, n, first_vertex, first_texture_coord, num_texture_coords;
  int tmp;

  if (! dest || ! src) {
    ERROR("missing arguments\n");
    return 1;
  }
    /* unlikely to happen */
  if (num_vertices == 0xffff) {
    error_dbg("the gsmhacking code assumes no mesh will ever have %u vertices\n", 0xffff);
    error_dbg("if the mesh really does have that many, and if it doesn't crash the game\n");
    error_dbg("send your GSM to furrycat and tell him you got this message\n");
    return 2;
  }

  /* first newly-created vertex */
  first_vertex = dest->num_vertices - num_vertices;

  *texture_coord_mapping = alloc_mapping(num_vertices);
  if (! *texture_coord_mapping) {
    ERROR("out of memory for texture co-ordinate mapping");
    return 3;
  }
  /* set default values in hash */
  initialise_mapping(*texture_coord_mapping);

  /* fill up the hash */
  n = 0;
  for (i = 0; i < num_vertices; i++) {
    vertex = dest->vertices[first_vertex + i];
    TRACE("vertex %p: position %u, uv %u\n", vertex, vertex->position_id, vertex->texture_coord_id);
    tmp = get_key_for_mapping(*texture_coord_mapping, vertex->texture_coord_id);
    if (tmp < -1) {
      ERROR("can't get mapping for texture co-ordinate in vertex %u\n", i);
      cleanup_mapping(*texture_coord_mapping);
      return 4;
    }
    if (tmp == -1 || tmp == 0xffff) {
      if (set_value_for_mapping(*texture_coord_mapping, n, vertex->texture_coord_id)) {
        ERROR("can't set mapping for texture co-ordinate %u to %u in new vertex %u\n", vertex->texture_coord_id, n, i);
        cleanup_mapping(*texture_coord_mapping);
        return 5;
      }
      n++;
    }
  }

  /* trash redundant entries in texture co-ordinate hash */
  for (i = (*texture_coord_mapping)->size - 1; i > 0; i--) {
    tmp = get_value_for_mapping(*texture_coord_mapping, i);
    if (tmp < -1) {
      ERROR("oh dear, texture co-ordinate mapping is bollocksed\n");
      cleanup_mapping(*texture_coord_mapping);
      return 6;
    }
    if (tmp > -1 && tmp < 0xffff) break;
    DEBUG("redundant entry %u of %u in texture co-ordinate mapping\n", i, (*texture_coord_mapping)->size);
    pop_mapping(*texture_coord_mapping, i);
  }
  sort_mapping_by_value(*texture_coord_mapping, SORT_MAPPING_NORMALISED);
  debug_mapping(*texture_coord_mapping, "texture_coord");

  /* make new texture_coords */
  num_texture_coords = (*texture_coord_mapping)->size;
  DEBUG("making %u new texture co-ordinate%s\n", num_vertices, (num_vertices == 1) ? "" : "s");
  new_texture_coords = alloc_more_texture_coords(num_texture_coords, dest->num_texture_coords, dest->texture_coords);
  if (! new_texture_coords) {
    ERROR("out of memory for new texture co-ordinates");
    cleanup_mapping(*texture_coord_mapping);
    return 7;
  }
  dest->texture_coords = new_texture_coords;
  first_texture_coord = dest->num_texture_coords;
  dest->num_texture_coords += num_texture_coords;
  /* add new ones from source */
  for (i = 0; i < num_texture_coords; i++) {
    new_texture_coord = new_texture_coords[i + first_texture_coord];
    tmp = get_value_for_mapping(*texture_coord_mapping, i);
    if (tmp < 0) {
      ERROR("can't get original texture co-ordinate from map index %u\n", i);
      return 8;
    }
    texture_coord = src->texture_coords[tmp];
    new_texture_coord->offset = 0x0fc0;
    new_texture_coord->u = texture_coord->u;
    new_texture_coord->v = texture_coord->v;
    TRACE("copying texture_coord %u as %u\n", (unsigned short) tmp, i);
  }

  /* last thing to do is set the right texture_coords in new vertices */
  for (i = first_vertex; i < dest->num_vertices; i++) {
    vertex = dest->vertices[i];
    /* assuming again the mapping won't let us down */
    tmp = get_key_for_mapping(*texture_coord_mapping, vertex->texture_coord_id) + first_texture_coord;
    vertex->texture_coord_id = (unsigned short) tmp;
    TRACE("setting texture co-ord to %u in vertex %u\n", (unsigned short) tmp, i);
  }

  cleanup_mapping(*texture_coord_mapping);

  return 0;
}

int copy_material_create_positions(struct gsm *dest, struct gsm *src, unsigned short num_vertices, mapping_t **position_mapping) {
  struct position *position, *new_position, **new_positions;
  struct vertex *vertex;
  unsigned short i, n, m, first_vertex, num_positions;
  mapping_t *bone_mapping;
  int tmp, bone_id;

  if (! dest || ! src) {
    ERROR("missing arguments\n");
    return 1;
  }
  /* unlikely to happen */
  if (num_vertices == 0xffff) {
    error_dbg("the gsmhacking code assumes no mesh will ever have %u vertices\n", 0xffff);
    error_dbg("if the mesh really does have that many, and if it doesn't crash the game\n");
    error_dbg("send your GSM to furrycat and tell him you got this message\n");
    return 2;
  }

  /* first newly-created vertex */
  first_vertex = dest->num_vertices - num_vertices;

  *position_mapping = alloc_mapping(num_vertices);
  if (! *position_mapping) {
    ERROR("out of memory for position mapping");
    return 3;
  }
  /* set default values in hash */
  initialise_mapping(*position_mapping);

  /* fill up the hash */
  n = 0;
  for (i = 0; i < num_vertices; i++) {
    vertex = dest->vertices[first_vertex + i];
    TRACE("vertex %p: position %u, uv %u\n", vertex, vertex->position_id, vertex->texture_coord_id);
    tmp = get_key_for_mapping(*position_mapping, vertex->position_id);
    if (tmp < -1) {
      ERROR("can't get mapping for position in vertex %u\n", i);
      cleanup_mapping(*position_mapping);
      return 4;
    }
    if (tmp == -1 || tmp == 0xffff) {
      if (set_value_for_mapping(*position_mapping, n, vertex->position_id)) {
        ERROR("can't set mapping for position %u to %u in new vertex %u\n", vertex->position_id, n, i);
        cleanup_mapping(*position_mapping);
        return 5;
      }
      n++;
    }
  }

  /* trash redundant entries in position hash */
  for (i = (*position_mapping)->size - 1; i > 0; i--) {
    tmp = get_value_for_mapping(*position_mapping, i);
    if (tmp < -1) {
      ERROR("oh dear, position mapping is bollocksed\n");
      cleanup_mapping(*position_mapping);
      return 6;
    }
    if (tmp > -1 && tmp < 0xffff) break;
    DEBUG("redundant entry %u of %u in position mapping\n", i, (*position_mapping)->size);
    pop_mapping(*position_mapping, i);
  }
  sort_mapping_by_value(*position_mapping, SORT_MAPPING_NORMALISED);
  debug_mapping(*position_mapping, "position");

  num_positions = (*position_mapping)->size;

  /* use the key of the mapping as bone_id and sort on index */
  /* we need the positions to be sorted by bone */
  for (i = 0; i < num_positions; i++) {
    tmp = get_value_for_mapping(*position_mapping, i);
    if (tmp < 0) goto fill_position_oops;
    if (tmp >= src->num_positions) {
      ERROR("source position %d out of range (max %u)\n", tmp, src->num_positions);
      if (tmp < src->num_positions + src->num_blended_positions) {
        ERROR("probably a blended position - may be supported later\n");
      }
      goto fill_position_oops;
    }
    position = src->positions[tmp];
    DEBUG("bone_mapping: position %u has bone %u\n", (unsigned short) tmp, position->bone_id);

    /* source may be a different model type */
    bone_id = reassign_bone_id(src->type, dest->type, position->bone_id);
    if (bone_id < 0) {
      ERROR("can't remap bone ID %u\n", position->bone_id);
      cleanup_mapping(*position_mapping);
      return 8;
    }
    if (bone_id >= dest->skeletons[0]->num_bones) {
      ERROR("reassigned bone %d (from %u) out of range\n", bone_id, position->bone_id);
      cleanup_mapping(*position_mapping);
      return 9;
    }

    if (set_key_for_mapping(*position_mapping, (unsigned short) tmp, bone_id)) {
fill_position_oops:
      ERROR("can't set bone mapping for position %u\n", i);
      cleanup_mapping(*position_mapping);
      return 10;
    }
  }
  sort_mapping_by_value(*position_mapping, SORT_MAPPING_DEFAULT);
  debug_mapping(*position_mapping, "position/bone");

  /* now we can make the new positions */
  if (copy_material_insert_positions(dest, src, *position_mapping, first_vertex)) {
    ERROR("can't insert new positions into positions array\n");
    cleanup_mapping(*position_mapping);
    return 11;
  }

  cleanup_mapping(*position_mapping);

  return 0;
}

/*
  for each set of new positions (grouped by bone) we insert the positions into 
  the position stream in the right place, edit subsequent position IDs and 
  assignments, then replace the bone_id side of the position mapping with the 
  new position ID
  we'll use this later on to adjust position_id for each vertex

  hang in there we've nearly done it
*/
int copy_material_insert_positions(struct gsm *dest, struct gsm *src, mapping_t *position_mapping, unsigned short first_vertex) {
  struct position *position, *new_position, **new_positions;
  struct blended_position *blended_position;
  struct assignment *assignment;
  struct vertex *vertex;
  unsigned short first_position, num_positions, num_extra_positions, *blend;
  unsigned short i, j, k, n, count, count_total, count_this_pass;
  int tmp;

  if (! dest || ! src || ! position_mapping) {
    ERROR("missing arguments\n");
    return 1;
  }

  num_positions = position_mapping->size;
  if (! num_positions) {
    ERROR("umm, no new positions!\n");
    return 2;
  }
  num_positions += dest->num_positions;
  first_position = dest->num_positions;

  /* fix vertices with blended positions */
  TRACE("updating vertices with blended positions\n");
  for (i = 0; i < dest->num_vertices; i++) {
    vertex = dest->vertices[i];
    if (vertex->position_id >= first_position) {
      vertex->position_id += position_mapping->size;
    }
  }
  dest->first_blended_position += position_mapping->size;

  /*
    we'll do it by hand instead of using alloc_more, as we need to add 
    several new sets of positions at arbitrary locations
  */
  new_positions = alloc_positions(num_positions);
  if (! new_positions) {
    ERROR("out of memory for new positions");
    return 3;
  }

  /* add positions for each assignment */
  n = count = num_extra_positions = 0;
  for (i = 0; i < dest->num_assignments; i++) {
    assignment = dest->assignments[i];

    /* original positions */
    TRACE("copying old positions with n at %u\n", n);
    for (j = 0; j < assignment->num_vertices; j++) {
      new_position = new_positions[n];
      position = dest->positions[assignment->first_vertex + j];
      TRACE1("copying position %u as %u\n", assignment->first_vertex + j, n);
      /* offset */
      new_position->offset = 0x0fc0;
      /* old_id for later processing */
      new_position->old_id = assignment->first_vertex + j;
      /* bone */
      new_position->bone_id = position->bone_id;
      /* co-ordinates */
      copy_vector(&new_position->local, &position->local);
      /* world position */
      copy_vector(&new_position->world, &position->world);

      /* next new position */
      n++;
    }

    /* new positions? */
    TRACE("adding new positions with n at %u\n", n);
    count_this_pass = count_total = 0;
    for (j = count; j < position_mapping->size; j++) {
      /* key is bone */
      tmp = get_key_in_mapping(position_mapping, j);
      if (tmp != assignment->bone_id) continue;

      /* value is source position */
      tmp = get_value_in_mapping(position_mapping, j);
      if (tmp >= src->num_positions) {
        ERROR("I don't believe it\n");
        return 4;
      }
      position = src->positions[tmp];
      new_position = new_positions[n];

      TRACE1("copying source position %d as %u\n", tmp, n);
      /* offset */
      new_position->offset = 0x0fc0;
      /* mark this as having no old ID */
      new_position->old_id = -1;
      /* bone */
      new_position->bone_id = position->bone_id;
      /* co-ordinates */
      copy_vector(&new_position->local, &position->local);
      /* world position */
      TRACE2("XXX: ignoring world position - strictly speaking this should be fixed\n");

      /* remember new position ID */
      if (set_key_for_mapping(position_mapping, tmp, n)) {
        ERROR("can't remember new position ID %u for source %d\n", n, tmp);
        return 5;
      }

      /* next new position */
      n++;
      /* next time round we start from this entry in the mapping */
      count++;
      /* this time round we added this many */
      count_this_pass++;
      /* total new positions added (not same as num_extra_positions yet) */
      count_total++;
    }

 /* 
    ``n'' is the index in new_positions
    ``count'' is the index of the position mapping we are currently on
    ``count_this_pass'' is the number of positions we added from this assignment
    ``count_total'' is the total number of new positions we have added
    ``num_extra_positions'' is the total number of new positions we have added
 */
    /* fix blended positions */
    TRACE("updating blended positions with n at %u\n", n);
    if (count_this_pass) {
      for (j = 0; j < dest->num_blended_positions; j++) {
        blended_position = dest->blended_positions[j];
        for (k = 0; k < blended_position->num_positions; k++) {
          blend = blended_position->positions[k];
          if (*blend >= assignment->first_vertex + assignment->num_vertices) {
            TRACE("count: %u: bumping blended position (%u, %u): %u up to %u\n", count, j, k, *blend, *blend + count_total);
            *blend += count_total;
            if (*blend >= num_positions) {
              ERROR("oops, bumped too far (%u > %u)\n", *blend, num_positions);
              return 6;
            }
          }
        }
      }
    }

    /* allow for any positions we previously inserted */
    TRACE1("adding %u to assignment %u first position to account for extra position%s\n", num_extra_positions, assignment->first_vertex, (num_extra_positions == 1) ? "" : "s");
    assignment->first_vertex += num_extra_positions;
    assignment->num_vertices += count_this_pass;

    num_extra_positions += count_this_pass;
  }

  /* now move the new positions into place */
  cleanup_positions(dest->num_positions, dest->positions);
  dest->num_positions = num_positions;
  dest->positions = new_positions;

  /* we need to update position IDs in vertices */
  TRACE("updating position IDs in vertices\n");
  for (i = 0; i < dest->num_vertices - position_mapping->size; i++) {
    vertex = dest->vertices[i];
    for (j = 0; j < dest->num_positions; j++) {
      position = dest->positions[j];
#if 0
      if (position->old_id == -1) {
        /* yeah so what? */
        SAY("Hey!  Position %u has old_id -1\n", j);
      }
#endif
      if (vertex->position_id == position->old_id) {
        TRACE("updating vertex %u position ID from %u to %u\n", i, vertex->position_id, j);
        vertex->position_id = j;
        position->offset = 0x0fc0;
        position->old_id == -2;
        break;
      }
    }
  }
  for (i = first_vertex; i < dest->num_vertices; i++) {
    j = i - first_vertex; /* new position index */
    vertex = dest->vertices[i];
    tmp = get_key_for_mapping(position_mapping, vertex->position_id);
    /* theorem: mapping is not same as new position index */
    TRACE("setting vertex %u (new vertex %u) position ID to %u (was %u)\n", i, j, (unsigned short) tmp, vertex->position_id);
    vertex->position_id = (unsigned short) tmp;
  }

  return 0;
}
