#include "gsmhacking.h"

struct triangle **alloc_triangles(unsigned short n) {
  return (struct triangle **) alloc_lots(n, "triangle", sizeof(struct triangle *), (void *(*)()) alloc_triangle);
}

struct triangle **alloc_more_triangles(unsigned short n, unsigned short num_triangles, struct triangle **triangles) {
  return (struct triangle **) alloc_more(n, "triangle", sizeof(struct triangle *), (void *(*)()) alloc_triangle, num_triangles, (void **) triangles, sizeof(struct triangle));
}

struct triangle *alloc_triangle() {
  return (struct triangle *) alloc_one("triangle", sizeof(struct triangle));
}

void cleanup_triangles(unsigned short num_triangles, struct triangle **triangles) {
  free_structs("triangle", num_triangles, (void *) triangles);
}

int find_triangles(struct gsm *gsm) {
  unsigned short i;
  unsigned int pos;
  struct triangle *triangle;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think triangles start at 0x%04x\n", pos);

  /* how many triangles are there? */
  if (read_short(gsm->fd, &gsm->num_triangles)) {
    ERROR("can't read number of triangles");
    return 2;
  }
  if (! gsm->num_triangles) {
    SAY("%s has no triangles\n", gsm->filename);
    return 3;
  }
  DEBUG("triangles: %u\n", gsm->num_triangles);

  /* allocate a big enough struct */
  gsm->triangles = alloc_triangles(gsm->num_triangles);
  if (! gsm->triangles) return 4;

  /* fill in the triangles */
  for (i = 0; i < gsm->num_triangles; i++) {
    triangle = gsm->triangles[i];
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    triangle->offset = pos;
    if (parse_triangle(gsm, i, triangle, gsm->num_triangles)) {
      ERROR("expected triangles id %u but got null\n", i);
      return 6;
    }
  }

  return 0;
}

int parse_triangle(struct gsm *gsm, unsigned short num, struct triangle *triangle, unsigned short total) {

  if (read_short(gsm->fd, &triangle->vertex1)) {
    ERROR("read failed parsing triangle 1st effective vertex\n");
    return 1;
  }
  if (triangle->vertex1 >= gsm->num_vertices) {
    ERROR("triangle %u vertex one %u out of range!\n", num, triangle->vertex1);
  }
  else gsm->vertices[triangle->vertex1]->num_triangles++;

  if (read_short(gsm->fd, &triangle->vertex2)) {
    ERROR("read failed parsing triangle 2nd effective vertex\n");
    return 2;
  }
  if (triangle->vertex2 >= gsm->num_vertices) {
    ERROR("triangle %u vertex one %u out of range!\n", num, triangle->vertex2);
  }
  else gsm->vertices[triangle->vertex2]->num_triangles++;

  if (read_short(gsm->fd, &triangle->vertex3)) {
    ERROR("read failed parsing triangle 3rd effective vertex\n");
    return 3;
  }
  if (triangle->vertex3 >= gsm->num_vertices) {
    ERROR("triangle %u vertex one %u out of range!\n", num, triangle->vertex3);
  }
  else gsm->vertices[triangle->vertex3]->num_triangles++;

  if (read_short(gsm->fd, &triangle->unknown)) {
    ERROR("read failed parsing triangle unknown\n");
    return 4;
  }

  TRACE("triangle %u at 0x%04x: (%u, %u, %u)\n", num, triangle->offset, triangle->vertex1, triangle->vertex2, triangle->vertex3);

  return 0;
}

int write_triangles(struct gsm *gsm) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing triangles: %u\n", gsm->num_triangles);
  if (write_short(gsm->out_fd, &gsm->num_triangles)) {
    ERROR("can't write number of texture co-ordinates");
    return 2;
  }

  for (i = 0; i < gsm->num_triangles; i++) {
    TRACE("writing triangle %u\n", i);
    if (write_short(gsm->out_fd, &gsm->triangles[i]->vertex1)) {
      ERROR("can't write triangle %u 1st vertex", i);
      return 3;
    }
    if (write_short(gsm->out_fd, &gsm->triangles[i]->vertex2)) {
      ERROR("can't write triangle %u 2nd vertex", i);
      return 4;
    }
    if (write_short(gsm->out_fd, &gsm->triangles[i]->vertex3)) {
      ERROR("can't write triangle %u 3rd vertex", i);
      return 5;
    }
    if (write_short(gsm->out_fd, &gsm->triangles[i]->unknown)) {
      ERROR("can't write triangle %u unknown", i);
      return 6;
    }
  }

  return 0;
}

/* return a triangle as (v0,v1,v2) */
char *get_triangle_as_positions(struct gsm *gsm, unsigned short triangle_id) {
  char *ret;
  int len;
  struct triangle *triangle;

  if (! gsm) {
    ERROR("no GSM\n");
    return "(BO,GU,S!)";
  }

  if (triangle_id >= gsm->num_triangles) {
    ERROR("triangle_id out of range\n");
    return "(BO,GU,S!)";
  }

  ret = (char *) malloc(19);
  if (! ret) {
    ERROR("out of memory for triangle description");
    return "(??,??,??)";
  }

  triangle = gsm->triangles[triangle_id];

  len = snprintf(ret, 19, "(% 4d, % 4d, % 4d)", get_position_by_vertex(gsm, triangle->vertex1), get_position_by_vertex(gsm, triangle->vertex2), get_position_by_vertex(gsm, triangle->vertex3));

  if (len == -1) {
    ERROR("snprintf() failed for triangle %d", triangle_id);
    free(ret);
    return "(??,??,??)";
  }

  return ret;
}

/*
  given a vertex ID, return 1 if it's the 1st vertex of the triangle, 2 if it's
  the second, 3 if it's the third, 0 if the vertex isn't in the triangle
*/
int get_triangle_vertex_by_id(struct gsm *gsm, unsigned short triangle_id, unsigned short vertex_id) {
  struct triangle *triangle;

  if (! gsm) {
    ERROR("no GSM!\n");
    return -4;
  }
  if (triangle_id >= gsm->num_triangles) {
    ERROR("triangle out of range\n");
    return -5;
  }
  if (vertex_id >= gsm->num_vertices) {
    ERROR("vertex out of range\n");
    return -6;
  }

  triangle = gsm->triangles[triangle_id];
  if (triangle->vertex1 == vertex_id) return 1;
  if (triangle->vertex2 == vertex_id) return 2;
  if (triangle->vertex3 == vertex_id) return 3;

  return 0;
}

/* like get_triangle_vertex_by_id() but using actual vertices ("positions") */
int get_triangle_vertex_by_position(struct gsm *gsm, unsigned short triangle_id, unsigned short position_id) {
  unsigned short i;
  int ret;

  if (! gsm) {
    ERROR("no GSM!\n");
    return -1;
  }
  if (triangle_id >= gsm->num_triangles) {
    ERROR("triangle out of range\n");
    return -2;
  }
  if (position_id >= gsm->num_positions) {
    ERROR("position out of range\n");
    return -3;
  }

  for (i = 0; i < gsm->num_vertices; i++) {
    if (gsm->vertices[i]->position_id != position_id) continue;
    ret = get_triangle_vertex_by_id(gsm, triangle_id, i);
    if (ret) return ret; /* handles -ve (error) and vertex id */
  }

  return 0;
}

int describe_triangles(struct gsm *gsm) {
  unsigned short i;
  struct triangle *triangle;
  char *tmp;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  SAY("Triangles: %u\n", gsm->num_triangles);
  if (gsm_verbose > 1) {
    for (i = 0; i < gsm->num_triangles; i++) {
      triangle = gsm->triangles[i];
      tmp = get_triangle_as_positions(gsm, i);
      SAY("  Triangle %04u at 0x%04x: (% 4u, % 4u, % 4u); positions: %s\n", i, triangle->offset, triangle->vertex1, triangle->vertex2, triangle->vertex3, tmp);
      free(tmp);
    }
  }

  return 0;
}

/* count triangles containing this position */
int get_position_reference_count(struct gsm *gsm, unsigned short position_id) {
  unsigned short i, j, count;
  int pos;
  struct triangle *triangle;

  if (! gsm) {
    ERROR("no GSM!\n");
    return -1;
  }
  if (position_id >= gsm->num_positions) {
    ERROR("position ID %u out of range\n", position_id);
    return -2;
  }

  TRACE("getting triangle count for position %u\n", position_id);
  count = 0;
  for (i = 0; i < gsm->num_triangles; i++) {
    triangle = gsm->triangles[i];
    /* first vertex */
    pos = get_position_by_vertex(gsm, triangle->vertex1);
    if (pos < 0) {
      ERROR("can't get position ID for vertex %u\n", triangle->vertex1);
      return -3;
    }
    /* second vertex */
    pos = get_position_by_vertex(gsm, triangle->vertex2);
    if (pos < 0) {
      ERROR("can't get position ID for vertex %u\n", triangle->vertex2);
      return -4;
    }
    /* third vertex */
    pos = get_position_by_vertex(gsm, triangle->vertex3);
    if (pos < 0) {
      ERROR("can't get position ID for vertex %u\n", triangle->vertex3);
      return -5;
    }
    if ((unsigned short) pos == position_id) count++;
  }

  return count;
}
