#include "gsmhacking.h"

struct vertex **alloc_vertices(unsigned short n) {
  return (struct vertex **) alloc_lots(n, "vertex", sizeof(struct vertex *), (void *(*)()) alloc_vertex);
}

struct vertex **alloc_more_vertices(unsigned short n, unsigned short num_vertices, struct vertex **vertices) {
  return (struct vertex **) alloc_more(n, "vertex", sizeof(struct vertex *), (void *(*)()) alloc_vertex, num_vertices, (void **) vertices, sizeof(struct vertex));
}

struct vertex *alloc_vertex() {
  struct vertex *ret;
  ret = alloc_one("vertex", sizeof(struct vertex));
  ret->count = -1;
}

void cleanup_vertices(unsigned short num_vertices, struct vertex **vertices) {
  free_structs("vertice", num_vertices, (void **) vertices);
}

int find_vertices(struct gsm *gsm) {
  unsigned short i;
  unsigned int pos;
  struct vertex *vertex;

  if (! gsm || ! gsm->fd) {
    ERROR("no file descriptor for gsm!\n");
    return 1;
  }

  pos = get_file_position(gsm->fd);
  TRACE("get_file_position({gsm->fd=%p}: 0x%04x\n", gsm->fd, pos);
  DEBUG("we think vertices start at 0x%04x\n", pos);

  /* how many vertices are there? */
  if (read_short(gsm->fd, &gsm->num_vertices)) {
    ERROR("can't read number of effective vertices");
    return 2;
  }
  if (! gsm->num_vertices) {
    SAY("%s has no effective vertices\n", gsm->filename);
    return 3;
  }
  DEBUG("effective vertices: %u\n", gsm->num_vertices);

  /* allocate a big enough struct */
  gsm->vertices = alloc_vertices(gsm->num_vertices);
  if (! gsm->vertices) return 4;

  /* fill in the vertices */
  for (i = 0; i < gsm->num_vertices; i++) {
    vertex = gsm->vertices[i];
    pos = get_file_position(gsm->fd);
    if (pos == -1) return 5;
    vertex->offset = pos;
    if (parse_vertex(gsm, i, vertex, gsm->num_vertices)) {
      ERROR("expected vertex id %u but got null\n", i);
      return 6;
    }
  }

  return 0;
}

int parse_vertex(struct gsm *gsm, unsigned short num, struct vertex *vertex, unsigned short total) {
  if (read_short(gsm->fd, &vertex->texture_coord_id)) {
    ERROR("read failed parsing vertex texture co-ordinate\n");
    return 1;
  }

  if (read_short(gsm->fd, &vertex->position_id)) {
    ERROR("read failed parsing vertex position\n");
    return 2;
  }
  if (vertex->position_id >= gsm->num_positions) {
    if (vertex->position_id >= gsm->num_positions + gsm->num_blended_positions) {
      ERROR("vertex %u out of position range\n", num);
      return 2;
    }
    TRACE2("vertex %u references blended position %u\n",  num, vertex->position_id - gsm->num_positions);
  }

  if (read_short(gsm->fd, &vertex->shared_triangle_list_id)) {
    ERROR("read failed parsing vertex shared triangle list\n");
    return 3;
  }

  if (read_short(gsm->fd, &vertex->hit_detect)) {
    ERROR("read failed parsing vertex hit detect\n");
    return 4;
  }

  TRACE("effective vertex %u at 0x%04x: position %02u, texture co-ord %02u, shared triangle list %02u, hit detect 0x%04x (%s)\n", num, vertex->offset, vertex->position_id, vertex->texture_coord_id, vertex->shared_triangle_list_id, vertex->hit_detect, hit_detect_location(vertex->hit_detect));

  return 0;
}

int write_vertices(struct gsm *gsm) {
  unsigned short i;

  if (! gsm || ! gsm->out_fd) {
    ERROR("no output file descriptor for gsm!\n");
    return 1;
  }

  DEBUG("writing effective vertics: %u\n", gsm->num_vertices);
  if (write_short(gsm->out_fd, &gsm->num_vertices)) {
    ERROR("can't write number of effective vertics");
    return 2;
  }

  for (i = 0; i < gsm->num_vertices; i++) {
    TRACE("writing vertex %u\n", i);
    if (write_short(gsm->out_fd, &gsm->vertices[i]->texture_coord_id)) {
      ERROR("can't write vertex %u texture co-ordinate", i);
      return 3;
    }
    if (write_short(gsm->out_fd, &gsm->vertices[i]->position_id)) {
      ERROR("can't write vertex %u position", i);
      return 4;
    }
    if (write_short(gsm->out_fd, &gsm->vertices[i]->shared_triangle_list_id)) {
      ERROR("can't write vertex %u shared triange list", i);
      return 5;
    }
    if (write_short(gsm->out_fd, &gsm->vertices[i]->hit_detect)) {
      ERROR("can't write vertex %u hit_detect", i);
      return 6;
    }
  }

  return 0;
}

int get_position_by_vertex(struct gsm *gsm, unsigned short vertex_id) {
  struct blended_position *blended_position;
  unsigned short i;
  int position_id;
  float weight;

  if (! gsm || gsm->num_vertices <= vertex_id) {
    ERROR("no vertex in gsm!\n");
    return -1;
  }
  position_id = gsm->vertices[vertex_id]->position_id;

  if (position_id < gsm->num_positions) return (int) position_id;

  /* blended position */
  position_id -= gsm->num_positions;
  TRACE("vertex %u maps blended position %u\n", vertex_id, position_id);

  /* for want of a better solution we the position with highest weight */
  blended_position = gsm->blended_positions[position_id];
  position_id = 0;
  weight = 0;
  for (i = 0; i < blended_position->num_positions; i++) {
    if (*blended_position->weights[i] > weight) {
      position_id = *blended_position->positions[i];
      weight = *blended_position->weights[i];
    }
  }

  return (int) position_id;
}

int get_texture_coord_by_vertex(struct gsm *gsm, unsigned short vertex_id) {
  if (! gsm || gsm->num_vertices <= vertex_id) {
    ERROR("no vertex in gsm!\n");
    return -1;
  }
  return (int) gsm->vertices[vertex_id]->texture_coord_id;
}

int get_vertex_by_count(struct gsm *gsm, int count) {
  unsigned short i;

  if (! gsm) {
    ERROR("no GSM!\n");
    return -1;
  }
  if (count < 0) {
    ERROR("bogus count %d\n", count);
    return -2;
  }

  for (i = 0; i < gsm->num_vertices; i++) {
    if (gsm->vertices[i]->count == count) return i;
  }

  return -1;
}

/* reassign the underlying position - needed for sorting when importing MS */
int set_position_for_vertices(struct gsm *gsm, unsigned short old_position_id, unsigned short new_position_id) {
  unsigned short i;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }
  if (old_position_id >= gsm->num_positions) {
    ERROR("old position out of range\n");
    return 2;
  }
  if (new_position_id >= gsm->num_positions) {
    ERROR("new position out of range\n");
    return 3;
  }

  DEBUG("for all vertices where position_id=%u, set position_id to %u\n", old_position_id, new_position_id);
  for (i = 0; i < gsm->num_vertices; i++) {
    if (gsm->vertices[i]->position_id == old_position_id) {
      TRACE("setting vertex %u's position_id to %u (was %u)\n",  i, new_position_id, old_position_id);
      gsm->vertices[i]->position_id = new_position_id;
    }
  }

  return 0;
}

int describe_vertices(struct gsm *gsm) {
  unsigned short i;
  struct vertex *vertex;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  SAY("Effective vertices: %u\n", gsm->num_vertices);
  for (i = 0; i < gsm->num_vertices; i++) {
    vertex = gsm->vertices[i];
    if (gsm_verbose > 1) {
      if (gsm->vertices[i]->position_id >= gsm->num_positions) {
        SAY("  Effective vertex %u at 0x%04x: blended vertex %02u, texture co-ord %02u, shared list %02u, hit detect %s\n", i, vertex->offset, vertex->position_id - gsm->num_positions, vertex->texture_coord_id, vertex->shared_triangle_list_id, hit_detect_location(vertex->hit_detect));
      }
      else {
        SAY("  Effective vertex %u at 0x%04x: vertex %02u, texture co-ord %02u, shared list %02u, hit detect %s\n", i, vertex->offset, vertex->position_id, vertex->texture_coord_id, vertex->shared_triangle_list_id, hit_detect_location(vertex->hit_detect));
      }
    }
    if (! vertex->num_triangles) SAY("    No triangles contain effective vertex %u\n", i);
  }

  return 0;
}

char *hit_detect_location(unsigned short rgb) {
  switch (rgb) {
    case 0x0000: return "NONE";
    case 0x7fff: return "explicitly NONE";
    case 0x2000: return "HEAD";
    case 0x4000: return "NECK";
    case 0x6000: return "CHEST";
    case 0x7c00: return "ABDOMEN";
    case 0x001f: return "PELVIS";
    case 0x2100: return "RIGHT SHOULDER";
    case 0x4200: return "RIGHT ELBOW";
    case 0x6300: return "RIGHT FOREARM";
    case 0x7fe0: return "RIGHT HAND";
    case 0x0108: return "LEFT SHOULDER";
    case 0x0210: return "LEFT ELBOW";
    case 0x0318: return "LEFT FOREARM";
    case 0x03ff: return "LEFT HAND";
    case 0x2008: return "RIGHT THIGH";
    case 0x4010: return "RIGHT KNEE";
    case 0x6018: return "RIGHT SHIN";
    case 0x7c1f: return "RIGHT FOOT";
    case 0x0100: return "LEFT THIGH";
    case 0x0200: return "LEFT KNEE";
    case 0x0300: return "LEFT SHIN";
    case 0x03e0: return "LEFT FOOT";
    default: return "UNKNOWN";
  }
}

int import_hit_detect(struct gsm *gsm) {
  struct gsm *src;
  unsigned short i;
  unsigned short len;
  int ret;

  if (! gsm) {
    ERROR("no GSM!\n");
    return 1;
  }

  if (! gsm->option->hit_detect) {
    ERROR("missing source\n");
    return 2;
  }

  src = alloc_gsm();
  if (! src) {
    ERROR("out of memory for source GSM");
    return 3;
  }

  src->filename = alloc_name(gsm->option->hit_detect, &len);
  if (! len) {
    ERROR("out of memory for source filename %s\n", gsm->option->hit_detect);
    return 4;
  }

  /* read in the source */
  DEBUG("reading source GSM\n");
  if (ret = read_gsm_file(src)) {
    if (ret > 0) ERROR("%s isn't a gsm file\n", src->filename);
    else ERROR("can't open %s\n", src->filename, ret);
    cleanup_gsm(src);
    return 5;
  }
  DEBUG("reading source GSM\n");
  if (parse_gsm(src)) {
    ERROR("can't parse mesh\n");
    cleanup_gsm(src);
    return 6;
  }

  DEBUG("crude geometry check\n");
  if (src->num_triangles != gsm->num_triangles) {
    ERROR("%s has %d triangles but %s has %d\n", src->filename, src->num_triangles, gsm->num_triangles);
    cleanup_gsm(src);
    return 7;
  }
  if (src->num_vertices != gsm->num_vertices) {
    ERROR("%s has %d vertices but %s has %d\n", src->filename, src->num_vertices, gsm->num_vertices);
    cleanup_gsm(src);
    return 8;
  }

  SAY("importing hit detection from %s\n", src->filename);
  for (i = 0; i < src->num_vertices; i++) {
    gsm->vertices[i]->hit_detect = src->vertices[i]->hit_detect;
  }

  cleanup_gsm(src);

  return 0;
}
